import json
import os
import pprint
import re
import subprocess
import time
import sys
import datetime
from io import StringIO
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.request import urlretrieve
import numpy as np
from onnx import ModelProto, TensorProto, load
from onnx.reference.op_run import to_array_extended
from ..reference import CReferenceEvaluator, from_array_extended


def save_for_benchmark_or_test(
    folder: str,
    test_name: str,
    model: ModelProto,
    inputs: List[np.ndarray],
    outputs: Optional[List[np.ndarray]] = None,
    data_set: int = 0,
) -> str:
    """
    Saves input, outputs on disk to later uses it as a backend test
    or a benchmark.

    :param folder: folder to save
    :param test_name: test name or subfolder
    :param model: model to save
    :param inputs: inputs of the node
    :param outputs: outputs of the node, supposedly the expected outputs,
        if not speficied, they are computed with the reference evaluator
    :param data_set: to have multiple tests with the same model
    :return: test folder
    """
    if outputs is None:
        ref = CReferenceEvaluator(model)
        outputs = ref.run(None, dict(zip([i.name for i in model.graph.input], inputs)))

    input_protos = [
        from_array_extended(inp, name)
        for name, inp in zip([i.name for i in model.graph.input], inputs)
    ]
    output_protos = [
        from_array_extended(out, name)
        for name, out in zip([i.name for i in model.graph.output], outputs)
    ]
    model_bytes = model.SerializeToString()

    path = os.path.join(folder, test_name)
    if not os.path.exists(path):
        os.makedirs(path)
    model_file = os.path.join(path, "model.onnx")
    with open(model_file, "wb") as f:
        f.write(model_bytes)

    sub_path = os.path.join(path, f"test_data_set_{data_set}")
    if not os.path.exists(sub_path):
        os.mkdir(sub_path)
    for i, t in enumerate(input_protos):
        with open(os.path.join(sub_path, f"input_{i}.pb"), "wb") as f:
            f.write(t.SerializeToString())
    for i, t in enumerate(output_protos):
        with open(os.path.join(sub_path, f"output_{i}.pb"), "wb") as f:
            f.write(t.SerializeToString())
    return path


class TestRun:
    """
    Loads a test saved by :func:`save_for_benchmark_or_test`.

    :param folder: test folder

    It has the following attributes:

    * `folder: str`
    * `proto: ModelProto`
    * `datasets: Dict[int, List[Tuple[int, np.array]]]`
    """

    def __init__(self, folder: str):
        self.folder = folder
        self._load()

    @staticmethod
    def _load_list(
        io_list: List[Tuple[int, str]], asarray: bool = True
    ) -> List[Tuple[int, TensorProto]]:
        new_list = []
        for i, name in io_list:
            with open(name, "rb") as f:
                content = f.read()
            pb = TensorProto()
            pb.ParseFromString(content)
            if asarray:
                new_list.append((i, to_array_extended(pb)))
            else:
                new_list.append((i, pb))
        return new_list

    def _load(self):
        "Loads the test."
        model = os.path.join(self.folder, "model.onnx")
        if not os.path.exists(model):
            raise FileNotFoundError(f"Unable to find {model!r}.")
        with open(model, "rb") as f:
            self.proto = load(f)

        datasets = []
        for sub in os.listdir(self.folder):
            if not sub.startswith("test_data_set_"):
                continue
            index = int(sub.replace("test_data_set_", ""))
            subf = os.path.join(self.folder, sub)
            ios = os.listdir(subf)
            inputs = []
            outputs = []
            for name in ios:
                fullname = os.path.join(subf, name)
                if name.startswith("input_"):
                    ii = int(os.path.splitext(name)[0][6:])
                    inputs.append((ii, os.path.join(subf, fullname)))
                elif name.startswith("output_"):
                    io = int(os.path.splitext(name)[0][7:])
                    outputs.append((io, os.path.join(subf, fullname)))
                else:
                    raise RuntimeError(f"Unexpected file {name!r} in {subf!r}.")

            inputs.sort()
            outputs.sort()
            inputs_pb = self._load_list(inputs)
            outputs_pb = self._load_list(outputs)
            datasets.append((index, (inputs_pb, outputs_pb)))

        self.datasets = dict(datasets)

    def __len__(self) -> int:
        "Returns the number of loaded datasets."
        return len(self.datasets)

    @property
    def input_names(self):
        "Returns the input names of the model."
        return [i.name for i in self.proto.graph.input]

    @property
    def output_names(self):
        "Returns the output names of the model."
        return [i.name for i in self.proto.graph.output]

    def test(
        self,
        f_build: Callable[[ModelProto], Any],
        f_run: Callable[[Any, Dict[str, np.array]], List[np.array]],
        index: int = 0,
        exc: bool = True,
        atol: float = 1e-5,
        rtol: float = 1e-5,
    ) -> Optional[Dict[str, Tuple[float, float, str]]]:
        """
        Runs the tests.

        :param f_build: function to call to build the inference class
        :param f_run: function to call to run the inference
        :param index: test index
        :param exc: raises an exception if the verification fails
        :param atol: absolute tolerance
        :param rtol: relative tolerance
        :return: list of results with discrepancies,
            the absolute error, the relative one and a reason for the failure
        """
        input_names = self.input_names
        output_names = self.output_names
        inputs, outputs = self.datasets[index]
        feeds = {}
        for ii, tensor in inputs:
            feeds[input_names[ii]] = tensor

        rt = f_build(self.proto)
        results = f_run(rt, feeds)

        d_outputs = dict(outputs)
        checks = {}
        for i, res in enumerate(results):
            name = output_names[i]
            expected = d_outputs[i]
            if expected.dtype != res.dtype:
                reason = (
                    f"Type mismatch for output {i}, {expected.dtype} != {res.dtype}"
                )
                checks[name] = (np.nan, np.nan, reason)
                continue
            if expected.shape != res.shape:
                reason = (
                    f"Shape mismatch for output {i}, {expected.shape} != {res.shape}"
                )
                checks[name] = (np.nan, np.nan, reason)
                continue
            expected64 = expected.astype(np.float64)
            res64 = res.astype(np.float64)
            diff = np.abs(res64 - expected64)
            max_diff = diff.max()
            rel_diff = (diff / np.maximum(expected, max(1e-15, atol))).max()

            if max_diff > atol:
                reason = f"Discrepancies for output {i}, atol={max_diff}"
                checks[name] = (max_diff, rel_diff, reason)
                continue
            if rel_diff > rtol:
                reason = f"Discrepancies for output {i}, rtol={rel_diff}"
                checks[name] = (max_diff, rel_diff, reason)
                continue

        if len(checks) == 0:
            return None
        if exc:
            raise AssertionError(
                f"Results do not match the expected value:"
                f"\n{pprint.pformat(checks)}"
            )
        return checks

    def bench(
        self,
        f_build: Callable[[ModelProto], Any],
        f_run: Callable[[Any, Dict[str, np.array]], List[np.array]],
        index: int = 0,
        warmup: int = 5,
        repeat: int = 10,
    ) -> Dict[str, Union[float, str, Any]]:
        """
        Runs the model on the given inputs.

        :param f_build: function to call to build the inference class
        :param f_run: function to call to run the inference
        :param index: test index to measure
        :param warmup: number of iterations to run before
            starting to measure the model
        :param repeat: number of iterations to measure
        :return: dictionary with many metrics,
            any metric endings with `"_time"` is a duration
        """
        str_type = {
            np.float16: "float16",
            np.float32: "float32",
            np.float64: "float64",
            np.int8: "int8",
            np.int16: "int16",
            np.int32: "int32",
            np.int64: "int64",
            np.uint8: "uint8",
            np.uint16: "uint16",
            np.uint32: "uint32",
            np.uint64: "uint64",
        }

        stats = {}
        begin = time.perf_counter()
        rt = f_build(self.proto)
        stats["build_time"] = time.perf_counter() - begin

        shapes = {}
        dtypes = {}
        input_names = self.input_names
        inputs = self.datasets[index][0]
        feeds = {}
        input_size = 0
        for ii, tensor in inputs:
            feeds[input_names[ii]] = tensor
            shapes[input_names[ii]] = tensor.shape
            dtypes[input_names[ii]] = str_type.get(tensor.dtype, str(tensor.dtype))
            input_size += np.prod(tensor.shape)
        stats["shapes"] = shapes
        stats["dtypes"] = dtypes
        stats["input_size"] = int(input_size)

        begin = time.perf_counter()
        for _ in range(warmup):
            f_run(rt, feeds)
        stats["warmup_time"] = time.perf_counter() - begin
        stats["warmup"] = warmup
        stats["name"] = self.folder
        stats["index"] = index

        ts = []
        for i in range(repeat):
            begin = time.perf_counter()
            f_run(rt, feeds)
            ts.append(time.perf_counter() - begin)

        stats["repeat"] = repeat
        stats["avg_time"] = float(np.array(ts).mean())
        stats["min_time"] = float(np.array(ts).min())
        stats["max_time"] = float(np.array(ts).max())

        ts.sort()
        if repeat > 4:
            stats["max1_time"] = ts[-2]
            stats["min1_time"] = ts[1]
            stats["avg1_time"] = float(np.array(ts[1:-1]).mean())
        return stats


def _run_cmd(args: List[str]) -> str:
    p = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env={"PYTHONPATH": ""}
    )
    st = StringIO()
    while True:
        output = p.stdout.readline().decode(errors="ignore")
        if output == "" and p.poll() is not None:
            break
        if output:
            out = output.rstrip()
            st.write(out + "\n")
    p.poll()
    p.stdout.close()
    return st.getvalue()


def _extract_version(out: str) -> str:
    reg = re.compile("Version: ([0-9][0-9.]*)")
    r = reg.findall(out)
    if not r:
        raise RuntimeError(f"Unable to find a version in\n{out}")
    return r[0]


def _display_conf(conf: Dict[str, Union[str, None]]) -> str:
    rows = []
    for k, v in sorted(conf.items()):
        if v is None:
            rows.append(k)
        else:
            rows.append(f"{k}=={v}")
    return " ".join(rows)


def bench_virtual(
    test_path: str,
    virtual_path: str,
    runtimes: Union[List[str], str] = "ReferenceEvaluator",
    index: int = 0,
    warmup: int = 5,
    repeat: int = 10,
    modules: Optional[List[Dict[str, Union[str, None]]]] = None,
    verbose: int = 0,
    save_as_dataframe: Optional[str] = None,
    filter_fct: Optional[Callable[[str, Dict[str, Union[str, None]]], bool]] = None,
) -> List[Dict[str, Union[float, Dict[str, Tuple[int, ...]]]]]:
    """
    Runs the same benchmark over different
    versions of the same packages in a virtual environment.

    :param test_path: test path
    :param virtual_path: path to the virtual environment
    :param runtimes: runtimes to measure
        (ReferenceEvaluation, CReferenceEvaluator, onnxruntime)
    :param index: test index to measure
    :param warmup: number of iterations to run before
        starting to measure the model
    :param repeat: number of iterations to measure
    :param modules: modules to install, example:
        `modules=[{"onnxruntime": "1.16.0", "onnx": "1.15.0"}]`
    :param filter_fct: to disable some of the configuration
        based on the runtime and the installed modules
    :param verbose: verbosity
    :param save_as_dataframe: saves as dataframe
    :return: list of statistics
    """
    exe = os.path.join(virtual_path, "bin", "python")
    if not os.path.exists(exe):
        if verbose > 0:
            print(f"[bench_virtual] create the virtual environment in {virtual_path!r}")
        out = _run_cmd([sys.executable, "-m", "venv", virtual_path])
        if verbose > 2:
            print(out)
        if not os.path.exists(exe):
            raise RuntimeError(f"The virtual environment was not created:\n{out}")
        get_pip = os.path.join(virtual_path, "get_pip.py")
        if not os.path.exists(get_pip):
            if verbose > 2:
                print("[bench_virtual] install pip")
            urlretrieve("https://bootstrap.pypa.io/get-pip.py", get_pip)
        out = _run_cmd([exe, get_pip])
        if verbose > 2:
            print(out)

    if modules is None:
        # ext = "https://github.com/sdpython/onnx-extended.git"
        modules = [
            {"onnxruntime": "1.16.0", "onnx": None, "onnx-extended": "0.2.2"},
            {
                "onnxruntime": "1.15.1",
                "onnx": None,
                "onnx-extended": "0.2.2",  # f"git+{ext}"},
            },
        ]
    if isinstance(runtimes, str):
        runtimes = [runtimes]

    # mandatory packages
    for name in ["setuptools", "wheel", "pybind11", "cython", "tomli", "packaging"]:
        out = _run_cmd([exe, "-m", "pip", "install", name])
        if verbose > 2:
            print(out)

    # packages defined by the user
    obs = []
    for i, conf in enumerate(modules):
        confs = _display_conf(conf)
        if verbose > 2:
            print("-------------------------------------------------------")
        if verbose > 0:
            print(
                f"[bench_virtual] {i+1}/{len(modules)} "
                f"{datetime.datetime.now():%H:%M:%S} {confs}"
            )

        for k, v in conf.items():
            # We first check if it installed.
            if verbose > 1:
                print(f"[bench_virtual] check version of {k}")
            out = _run_cmd([exe, "-m", "pip", "show", k])
            if verbose > 2:
                print(out)
            if "Package(s) not found" in out:
                version = None
            else:
                version = _extract_version(out)
            if verbose > 1:
                print(f"[bench_virtual] found version of {k}: {version}")
            if version == v:
                continue

            # If the version is different, let's uninstall first.
            if verbose > 1:
                print(f"[bench_virtual] uninstall {k}")
            out = _run_cmd([exe, "-m", "pip", "uninstall", "-y", k])
            if verbose > 2:
                print(out)

            # Let's install the new one.
            if verbose > 1:
                print(f"[bench_virtual] install {k}: {v or 'upgrade'}")
            if v is None:
                out = _run_cmd([exe, "-m", "pip", "install", k, "--upgrade"])
                if verbose > 2:
                    print(out)
            elif v.startswith("git"):
                out = _run_cmd([exe, "-m", "pip", "install", v])
                if verbose > 2:
                    print(out)
            else:
                out = _run_cmd([exe, "-m", "pip", "install", f"{k}=={v}"])
                if verbose > 2:
                    print(out)
            if verbose > 1:
                print(f"[bench_virtual] check {k} is installed")
            out = _run_cmd(
                [exe, "-c", f"import {k.replace('-', '_')} as m;print(m.__file__)"]
            )
            if verbose > 2:
                print(out)
            if "Error" in out:
                raise RuntimeError(out)

        for rt in runtimes:
            if filter_fct is not None and not filter_fct(rt, conf):
                continue
            if verbose > 1:
                print(f"[bench_virtual] run with {rt}")
            cmd = [
                exe,
                "-m",
                "onnx_extended.tools.run_onnx_main",
                "-p",
                test_path,
                "-r",
                str(repeat),
                "-w",
                str(warmup),
                "-e",
                rt,
            ]
            out = _run_cmd(cmd)
            if "Traceback" in out:
                raise RuntimeError(out)
            try:
                js = json.loads(out)
            except json.decoder.JSONDecodeError as e:
                raise RuntimeError(f"Unable to decode {out!r}") from e
            if verbose > 2:
                print("[bench_virtual] final results")
                print(js)
            js["conf"] = confs
            js["conf_dict"] = conf
            js["cmd"] = " ".join(cmd)
            js["runtime"] = rt
            obs.append(js)

    if save_as_dataframe:
        import pandas

        data = []
        for o in obs:
            r = {}
            for key, value in o.items():
                if value is None:
                    continue
                if key == "bench":
                    for k, v in value.items():
                        r[f"b_{k}"] = v
                elif key == "test":
                    for k, v in value.items():
                        r[f"t_{k}"] = v
                elif key == "conf_dict":
                    for k, v in value.items():
                        r[f"v_{k}"] = v
                else:
                    r[key] = value
            data.append(r)

        df = pandas.DataFrame(data)
        df.to_csv(save_as_dataframe, index=False)
        return df
    return obs



import numpy


class _ClassifierCommon:
    """
    Labels strings are not natively implemented in C++ runtime.
    The class stores the strings labels, replaces them by
    integer, calls the C++ codes and then replaces them by strings.
    """

    @staticmethod
    def _post_process_predicted_label(label, scores, classlabels_int64s_string):
        """
        Replaces int64 predicted labels by the corresponding
        strings.
        """
        if classlabels_int64s_string is not None and len(classlabels_int64s_string) > 0:
            new_label = []
            no_array = False
            for i in label:
                if i >= len(classlabels_int64s_string):
                    new_label.append(None)
                    no_array = True
                else:
                    new_label.append(classlabels_int64s_string[i])
            if no_array:
                return new_label, scores
            return numpy.array(new_label), scores
        return label, scores



import os
import textwrap
from typing import List
from ... import _get_ort_ext_libs


def get_ort_ext_libs() -> List[str]:
    """
    Returns the list of libraries implementing new simple
    :epkg:`onnxruntime` kernels implemented for the
    :epkg:`CUDAExecutionProvider`.
    """
    libs = _get_ort_ext_libs(os.path.dirname(__file__))
    return [lib for lib in libs if "cuda_cuda" not in lib]


def documentation() -> List[str]:
    """
    Returns a list of rst string documenting every implemented kernels
    in this subfolder.
    """
    return list(
        map(
            textwrap.dedent,
            [
                """
    onnx_extented.ortops.tutorial.cuda.CustomGemm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    It calls CUDA library for Gemm :math:`\\alpha A B + \\beta C`.

    **Provider**
    
    CUDAExecutionProvider
    
    **Inputs**
    
    * A (T): tensor of type T
    * B (T): tensor of type T
    * C (T): tensor of type T
    * D (T): tensor of type T
    * E (T): tensor of type T

    **Outputs**

    * Z (T): :math:`\\alpha A B + \\beta C`

    **Constraints**

    * T: float, float16, bfloat16
    """
            ],
        )
    )



from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
from onnx import (
    AttributeProto,
    FunctionProto,
    GraphProto,
    ModelProto,
    NodeProto,
    SparseTensorProto,
    TensorProto,
    TypeProto,
    ValueInfoProto,
)
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_opsetid,
    set_model_props,
)
from onnx.shape_inference import infer_shapes
from onnx.version_converter import convert_version
from ...reference import CReferenceEvaluator, from_array_extended


def _get_shape(ttype: TypeProto) -> Optional[Tuple[Union[None, str, int], ...]]:
    """
    Returns the shape of a TypeProto.

    :param name: instance of TypeProto
    :return: None if unknown or a tuple
    """
    if not ttype.tensor_type:
        return None
    shape = ttype.tensor_type.shape
    res = [(d.dim_value if d.dim_value else d.dim_param) for d in shape.dim]
    return tuple(res)


class NodeKind(Enum):
    """
    Node kind.
    """

    UNDEFINED = 0
    INITIALIZER = 1
    SPARSE_INITIALIZER = 3
    INPUT = 4
    OUTPUT = 8
    NODE = 16


class Node:
    """
    Defines a node in the graph.
    It can be an iniatialier or a node.
    """

    def __init__(
        self,
        index: int,
        parent: "Graph",
        proto: Union[TensorProto, NodeProto, ValueInfoProto, str],
        kind: Optional[NodeKind] = None,
    ):
        if not isinstance(proto, (TensorProto, NodeProto, ValueInfoProto, str)):
            raise TypeError(f"Unexpected type {type(proto)} for proto.")
        if isinstance(proto, NodeProto) and proto.op_type == "Constant":
            if kind is None:
                kind = NodeKind.NODE
            elif kind != NodeKind.NODE:
                raise ValueError(f"Unexpected kind {kind!r} for a constant.")
            missing = True
            for att in proto.attribute:
                if att.name in {
                    "sparse_value",
                    "value",
                    "value_float",
                    "value_floats",
                    "value_int",
                    "value_ints",
                    "value_string",
                    "value_strings",
                }:
                    missing = False
                    break
            if missing:
                raise ValueError(f"Unexpected constant node {proto}.")
        if isinstance(proto, NodeProto):
            if kind is None:
                kind = NodeKind.NODE
            elif kind != NodeKind.NODE:
                raise ValueError(
                    f"Unexpected kind {kind!r} for a node type " f"{proto.op_type!r}."
                )
        if isinstance(proto, TensorProto):
            if kind is None:
                kind = NodeKind.INITIALIZER
            elif kind != NodeKind.INITIALIZER:
                raise ValueError(f"Unexpected kind {kind!r} for an initializer.")
            if not hasattr(proto, "name") or not proto.name:
                raise AttributeError("Attribute 'name' is missing for an initializer.")
        if kind is None:
            raise ValueError(
                f"kind is None and cannot specified for type(proto)={type(proto)}."
            )
        self.index = index
        self.proto = proto
        self.parent = parent
        self.kind = kind

    @property
    def name(self):
        "Returns the name if node is a NodeProto, None otherwise."
        if isinstance(self.proto, NodeProto):
            return self.proto.name
        return None

    def match(self, pattern: Optional[Dict[str, str]]) -> bool:
        """
        Checks if a node match the proposed pattern.

        :param pattern: a node matches the pattern `{"name": "node_name"}`
            if its node is equal to `'node_name'`
        :return: match
        """
        if pattern is None:
            return False
        for k, v in pattern.items():
            if k == "name":
                return v == self.name
            raise ValueError(f"Unexpected pattern key k={k!r}, v={v!r}")
        return False

    def get_tensor(self) -> TensorProto:
        "Returns the value of the"
        if self.is_node:
            if self.op_type == "Constant":
                model = CReferenceEvaluator(self.proto)
                arr = model.run(None, {})[0]
                return from_array_extended(arr, name=self.outname)
            raise NotImplementedError(
                f"{self.outname!r} is a constant obtained from other constant. "
                f"This case is not implemented yet."
            )
        if self.is_input:
            raise RuntimeError(f"{self.outname!r} is an input not a tensor.")
        if self.is_output:
            raise RuntimeError(f"{self.outname!r} is an output not a tensor.")
        return self.proto

    @property
    def outname(self):
        "Returns the output name."
        if len(self.outputs) != 1:
            raise RuntimeError(f"The node has more than one output: {self.outputs}.")
        return self.outputs[0]

    def __str__(self) -> str:
        if self.is_node:
            if self.op_type == "Constant":
                t = self.get_tensor()
                shape = tuple(t.dims)
                stype = f"{t.data_type}:{shape}"
                return (
                    f"{self.__class__.__name__}({self.index}, "
                    f"<parent>, <{self.op_type}>) "
                    f"[{stype}] -> [{','.join(self.outputs)}]"
                )
            return (
                f"{self.__class__.__name__}({self.index}, <parent>, <{self.op_type}>) "
                f"[{','.join(self.inputs)}] -> [{','.join(self.outputs)}]"
            )
        if isinstance(self.proto, TensorProto):
            shape = tuple(self.proto.dims)
            stype = f"{self.proto.data_type}:{shape}"
            return (
                f"{self.__class__.__name__}({self.index}, <parent>, "
                f"kind={self.kind}) "
                f"[{stype}] -> [{','.join(self.outputs)}]"
            )
        shape = _get_shape(self.proto.type)
        stype = f"{self.proto.type.tensor_type.elem_type}:{shape}"
        return (
            f"{self.__class__.__name__}({self.index}, <parent>, "
            f"kind={self.kind}) "
            f"[{stype}] -> [{','.join(self.outputs)}]"
        )

    @property
    def is_node(self) -> bool:
        "True if a NodeProto."
        return isinstance(self.proto, NodeProto)

    @property
    def is_input(self) -> bool:
        "True if an input"
        if (
            isinstance(self.proto, (str, ValueInfoProto))
            and self.kind == NodeKind.INPUT
        ):
            return True
        return False

    @property
    def is_output(self) -> bool:
        "True if an output"
        if (
            isinstance(self.proto, (str, ValueInfoProto))
            and self.kind == NodeKind.OUTPUT
        ):
            return True
        return False

    @property
    def is_initializer(self) -> bool:
        "True if inititializer"
        return isinstance(self.proto, TensorProto)

    @property
    def is_sparse_initializer(self) -> bool:
        "True if inititializer"
        return isinstance(self.proto, SparseTensorProto)

    @property
    def op_type(self) -> str:
        "Returns the node type."
        if self.is_input:
            # It is an input.
            return "input"
        if self.is_output:
            # It is an output.
            return "output"
        return self.proto.op_type if self.is_node else "initializer"

    def is_constant(self) -> bool:
        """
        True if operator Constant or initializer or a Constant as
        an output of an operator taking only constants.
        """
        if self.is_node:
            if self.proto.op_type == "Constant":
                return True
            return self._is_constant()
        return not (self.is_input or self.is_output)

    def _is_constant(self) -> bool:
        "Tells if a node is a constant or operate on constants."
        # This function is recursive and its results may
        # be cached for better performance.
        for i in self.inputs:
            if i not in self.parent.index_output:
                raise RuntimeError(f"Unable to find output {i!r} in the graph.")
            ni = self.parent.index_output[i]
            if not ni.is_constant():
                return False
        return True

    @property
    def inputs(self) -> List[str]:
        "Input names"
        if self.is_node:
            return self.proto.input
        return []

    @property
    def outputs(self) -> List[str]:
        "Output names"
        if self.is_node:
            return self.proto.output
        return [self.proto.name]

    def create_with_new_values(self, new_tensor: TensorProto) -> "Node":
        "Creates an iniatializer or a node Constant based on the new value."
        if self.is_node:
            new_name = self.parent.generate_name(new_tensor.name)
            node = make_node("Constant", [], [new_name], value=new_tensor)
            return Node(None, self.parent, node, NodeKind.NODE)
        # initializer
        new_tensor.name = self.parent.generate_name(new_tensor.name)
        return Node(None, self.parent, new_tensor, NodeKind.INITIALIZER)

    def getattr(
        self, name: str, astype: Optional[type] = None, has_default: bool = False
    ) -> Any:
        """
        Retrieves a specific attribute and extracts its value if
        *astype* is not None.

        :param name: attribute name
        :param astype: cast the attribute into this type
        :param has_default: if the parameter has a default value,
            the method returns None if the attribute is not found
        :return: the value of the attribute or an AttributeProto
            if *astype* is None
        """
        if not self.is_node:
            raise AttributeError(
                f"This node does not store an ONNX node but {self.op_type!r}."
            )
        proto = None
        for att in self.proto.attribute:
            if att.name == name:
                proto = att
                break
        if proto is None:
            if has_default:
                return None
            raise AttributeError(
                f"Unable to find attribute {name!r} in node type {self.op_type!r}."
            )
        if astype is None:
            return proto
        if astype is int:
            return proto.i
        raise NotImplementedError(
            f"Attribute name {name!r} for node {self.op_type!r} "
            f"cannot be cast into {astype!r}. Attribute is {proto}."
        )


class NodeWithSubGraph(Node):
    """
    A node with a subgraphs (If, Loop, Scan, ...).
    """

    def __init__(self, index: int, parent: "Graph", proto: NodeProto):
        if not isinstance(proto, NodeProto):
            raise TypeError(f"proto is not a NodeProto but {type(proto)}.")
        Node.__init__(self, index, parent, proto)
        self.subgraphs = {}
        for att in proto.attribute:
            if att.data_type == AttributeProto.GRAPH:
                self.subgraphs[att.name] = Graph(att.g)
        if len(self.subgraphs) == 0:
            raise ValueError(f"A node type {self.proto.op_type!r} has no subgraph.")

    @property
    def inputs(self) -> List[str]:
        raise NotImplementedError(
            f"It should return the implicit inputs for node type {self.op_type!r}."
        )


class NodeSet:
    """
    Defines a set of nodes.
    """

    def __init__(self, nodes: List[Node]):
        self.nodes = nodes

    def __len__(self) -> int:
        return len(self.nodes)

    def __iter__(self) -> Iterable[Node]:
        for n in self.nodes:
            yield n


class Graph:
    """
    A GraphProto, FunctionProto or ModelProto.
    """

    @staticmethod
    def node_or_node(proto: Union[TensorProto, NodeProto, ValueInfoProto, str]):
        if isinstance(proto, (TensorProto, ValueInfoProto, str)):
            return Node
        for att in proto.attribute:
            if att.type == AttributeProto.GRAPH:
                return NodeWithSubGraph
        return Node

    def _get_nodes(self, graph: Union[GraphProto, FunctionProto]) -> List[Node]:
        """
        Returns the ordered list of nodes.
        """
        nodes = []
        if isinstance(graph, GraphProto):
            for inp in graph.input:
                nodes.append(Node(len(nodes), self, inp, NodeKind.INPUT))
            for init in graph.initializer:
                nodes.append(Node(len(nodes), self, init, NodeKind.INITIALIZER))
            for init in graph.sparse_initializer:
                nodes.append(Node(len(nodes), self, init, NodeKind.SPARSE_INITIALIZER))
        else:
            for inp in graph.input:
                nodes.append(Node(len(nodes), self, inp, NodeKind.INPUT))
        for node in graph.node:
            nodes.append(
                Graph.node_or_node(node)(len(nodes), self, node, NodeKind.NODE)
            )
        if isinstance(graph, GraphProto):
            for inp in graph.output:
                nodes.append(Node(len(nodes), self, inp, NodeKind.OUTPUT))
        else:
            for inp in graph.output:
                nodes.append(Node(len(nodes), self, inp, NodeKind.OUTPUT))

        return nodes

    def __init__(self, proto: Union[FunctionProto, GraphProto, ModelProto]):
        self.proto = proto
        if isinstance(proto, ModelProto):
            graph = proto.graph
            if len(proto.functions) > 0:
                raise NotImplementedError(
                    "Class Graph does not handle model included functions yet."
                )
            self.functions: Dict[Tuple[str, str], FunctionProto] = {
                (f.domain, f.name): f for f in proto.functions
            }

            # retrieve all shapes
            p2 = infer_shapes(proto)
            values = p2.graph.value_info
            shapes: Dict[str, TypeProto] = {}
            for o in proto.graph.input:
                if o.name not in shapes:
                    shapes[o.name] = o.type
            for o in proto.graph.output:
                if o.name not in shapes:
                    shapes[o.name] = o.type
            for value in values:
                shapes[value.name] = value.type
            self.shapes: Dict[str, TypeProto] = shapes

        else:
            graph = proto
            self.shapes: Dict[str, TypeProto] = None
            self.functions: Dict[Tuple[str, str], FunctionProto] = {}

        self.nodes = self._get_nodes(graph)
        self.opsets: Dict[str, int] = {}
        self._complete_init()

    def _complete_init(self):
        self.graph_inputs: List[str] = []
        self.graph_outputs: List[str] = []
        self.removed: Set[str] = set()
        self.index_input: Dict[str, List[Node]] = {}
        self.index_output: Dict[str, Node] = {}
        self.nodes_added: Dict[int, Node] = {}
        self.nodes_sets: Dict[int:NodeSet] = {}
        self.generated_names: Set[str] = set()
        self.generated_node_names: Set[str] = set()
        self.new_index: int = len(self.nodes)

        for node in self.nodes:
            self._complete_init_node(node)

    def _complete_init_node(self, node):
        if node.is_input:
            self.graph_inputs.append(node.outputs[0])
        elif node.is_output:
            self.graph_outputs.append(node.outputs[0])
        if node.name not in ("", None):
            self.generated_node_names.add(node.name)
        for i in node.inputs:
            if i not in self.index_input:
                self.index_input[i] = []
            self.index_input[i].append(node)
            if i != "":
                self.generated_names.add(i)
        for i in node.outputs:
            self.index_output[i] = node
            if i != "":
                self.generated_names.add(i)

    def get_shape(self, name: str) -> Optional[Tuple[Union[None, str, int], ...]]:
        """
        Returns the shape of a result.

        :param name: name of the result
        :return: None if unknown or a tuple
        """
        if name not in self.shapes:
            return None
        ttype = self.shapes[name]
        return _get_shape(ttype)

    def _exists_name(self, name):
        if name in self.index_input:
            return True
        if name in self.index_output:
            return True
        if name in self.graph_inputs:
            return True
        if name in self.generated_names:
            return True
        return False

    def _exists_node_name(self, name):
        if name in self.generated_node_names:
            return True
        return False

    def generate_name(self, prefix: str = "new") -> str:
        """
        Generates a name which is not used for any existing result in the graph.

        :param prefix: prefix to use for the new name,
            next tries will be ``<prefix>_1``, ``<prefix>_2``, ...
        :return: new name
        """
        suggestion = prefix
        i = 0
        while self._exists_name(suggestion):
            i += 1
            suggestion = f"{prefix}_{i}"
        self.generated_names.add(suggestion)
        return suggestion

    def generate_node_name(self, prefix: str = "new") -> str:
        """
        Generates a node name which is not used for
        any existing node in the graph.

        :param prefix: prefix to use for the new name,
            next tries will be ``<prefix>_1``, ``<prefix>_2``, ...
        :return: new name
        """
        suggestion = prefix
        i = 0
        while self._exists_node_name(suggestion):
            i += 1
            suggestion = f"{prefix}_{i}"
        self.generated_node_names.add(suggestion)
        return suggestion

    def get_node_producer(self, name: str) -> Node:
        """
        Returns the node producing the output *name*.

        :param name: output name to check
        :return: Node producing the output *name* or None if it is an input.
        """
        if name not in self.index_input:
            raise ValueError(f"Unable to find any node producing output {name!r}.")
        return self.index_output[name]

    def get_opsets(self) -> Dict[str, int]:
        """
        Returns the opsets available registered for ever domain
        in the model.
        """
        if not isinstance(self.proto, ModelProto):
            raise TypeError(
                f"The graph does not represent a ModelProto but {type(self.proto)}."
            )
        res = {}
        for op in self.proto.opset_import:
            res[op.domain] = op.version
        res.update(self.opsets)
        return res

    def get_opset(self, domain: str = "") -> int:
        """
        Returns the opset for a specific domain.

        :param domain: domain
        :return: model opset
        """
        if not isinstance(self.proto, ModelProto):
            raise TypeError(
                f"The graph does not represent a ModelProto but {type(self.proto)}."
            )
        for op in self.proto.opset_import:
            if op.domain == domain:
                return op.version
        if domain in self.opsets:
            return self.opsets[domain]
        raise RuntimeError(f"Domain {domain!r} is not part the the model.")

    def is_constant(self, name: str) -> bool:
        """
        Tells if output *name* is constant.

        :param name: result name
        :return: True if constant
        """
        if name in self.graph_inputs:
            return False
        node = self.index_output[name]
        return node.is_constant()

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}(...) "
            f"[{','.join(self.graph_inputs)}] -> [{','.join(self.graph_outputs)}]"
        )

    def __len__(self) -> int:
        "Returns the number of nodes"
        return len(self.nodes) + len(self.nodes_added) - len(self.removed)

    def __getitem__(self, index: int) -> Node:
        """
        Returns a node at a specific index.
        """
        if index < len(self.nodes):
            node = self.nodes[index]
            if node is None:
                if index not in self.nodes_added:
                    raise IndexError(f"Unable to find node index {index}.")
            else:
                return node
        node = self.nodes_added[index]
        if node is None:
            raise IndexError(f"This node was probably reduced {index}.")
        return node

    def __iter__(self) -> Iterable[Node]:
        "Iterates on nodes or initializer."
        for index, node in enumerate(self.nodes):
            if node is None or node.index in self.removed:
                if index in self.nodes_sets:
                    for n in self.nodes_sets[index]:
                        yield n
                continue
            yield node

    def replace_nodes(
        self,
        indices: Union[int, List[int]],
        new_nodes: Union[NodeProto, List[NodeProto]],
        new_opsets: Optional[Dict[str, int]] = None,
    ) -> NodeSet:
        """
        Replaces a node index

        :param indices: index or list of indices to replace
        :param new_nodes: node or list of nodes to add
        :param new_opsets: new opet versions
        :return: added nodes

        By default, the nodes are inserted at position `indices[-1]`.
        It ensures the inputs of the new nodes were already computed.
        However, it does not ensure that every intermediate node
        between the first and the last removed nodes can be computed.
        Sorting the nodes is needed in that. This function does not do
        that.
        """
        if isinstance(new_nodes, NodeProto):
            new_nodes = [new_nodes]
        if isinstance(indices, int):
            indices = [indices]

        removed = []
        for index in indices:
            if index <= len(self.nodes):
                node = self.nodes[index]
                if node is None:
                    raise RuntimeError(f"Node index {index} was already removed.")
                removed.append((index, self.nodes[index]))
                self.nodes[index] = None
            elif index not in self.nodes_added:
                raise RuntimeError(
                    f"Node index {index} does not exists or was already removed."
                )
            if index in self.removed:
                raise RuntimeError(f"Node index {index} was already removed.")

        kind = None
        for index, node in removed:
            if kind is None:
                kind = node.kind
            elif node.kind is not None:
                if node.kind != kind:
                    kind = NodeKind.UNDEFINED
            self.removed.add(index)
            for i in node.inputs:
                new_input = [n for n in self.index_input[i] if n.index != index]
                self.index_input[i] = new_input
            for o in node.outputs:
                del self.index_output[o]
            if node.is_input:
                ni = node.outputs[0]
                if ni not in self.graph_inputs:
                    raise RuntimeError(
                        f"Removing node {node} but it was not "
                        f"found in self.graph_inputs."
                    )
                del self.graph_inputs[self.graph_inputs.index(ni)]
            elif node.is_output:
                ni = node.outputs[0]
                if ni not in self.graph_outputs:
                    raise RuntimeError(
                        f"Removing node {node} but it was not "
                        f"found in self.graph_outputs."
                    )
                del self.graph_outputs[self.graph_outputs.index(ni)]

        if kind == NodeKind.UNDEFINED:
            kind = None
        nodes = []
        new_indices = []
        for node in new_nodes:
            n = Node(self.new_index, self, node, kind=kind)
            self._complete_init_node(n)
            self.nodes_added[self.new_index] = n
            new_indices.append(self.new_index)
            self.new_index += 1
            nodes.append(n)

        nodes_set = NodeSet(nodes)
        new_pos = indices[-1]
        if new_pos in self.nodes_sets:
            raise NotImplementedError(
                f"Nodes were already added at position {new_pos}. "
                f"This conflicts is not yet handled."
            )
        self.nodes_sets[new_pos] = nodes_set
        if new_opsets is not None:
            self.opsets.update(new_opsets)
        return nodes_set

    def simplify(self, remove_unused: bool = True) -> "Graph":
        """
        Stores every node into nodes.
        Removes unused nodes.

        :param remove_unused: removes unused nodes as well,
            see :meth:`remove_unused_nodes`
        :return: self
        """
        if (
            len(self.removed) == 0
            and len(self.nodes_added) == 0
            and len(self.nodes_sets) == 0
        ):
            # Nothing to do.
            return

        self.nodes = list(self)
        self._complete_init()
        for i, node in enumerate(self.nodes):
            node.index = i
        if remove_unused:
            self.remove_unused_nodes()
        return self

    def remove_unused_nodes(self):
        """
        Removes unused nodes, a node with only unused outputs.

        :return: removed nodes
        """
        total_remove = []
        while True:
            to_remove = []
            for node in self:
                rem = 0
                for name in node.outputs:
                    if (
                        name not in {"", None}
                        and name not in self.index_input
                        and name not in self.graph_outputs
                    ):
                        rem += 1
                if rem < len(node.outputs):
                    # one outputs is used
                    continue
                to_remove.append(node)
                self.removed.add(node.index)

            if len(to_remove) == 0:
                break

            total_remove.extend(to_remove)
            self.simplify(False)
        return total_remove

    def upgrade_opsets(self, new_opsets: Dict[str, int]):
        """
        Upgrades the models to a newer opsets.

        :param new_opsets: dictionary { domain: new version }
        """
        if not isinstance(self.proto, ModelProto):
            raise RuntimeError(
                f"Upgrading a model only works on a ModelProto not {type(self.proto)}."
            )
        if len(new_opsets) != 1 or "" not in new_opsets:
            raise RuntimeError(
                f"Upgrade an opset only work for domain '' "
                f"but new_opsets={new_opsets}."
            )
        new_proto = convert_version(self.proto, new_opsets[""])
        self.proto = new_proto
        self.nodes = self._get_nodes(self.proto.graph)
        self._complete_init()

    def add_functions(self, protos: Iterable[FunctionProto]):
        """
        Adds functions to the graph when it is exported to ONNX.

        :param protos: enumerate of FunctionProto
        """
        for proto in protos:
            if not isinstance(proto, FunctionProto):
                raise TypeError(f"Unexpected type {type(proto)} for a function.")
            key = proto.domain, proto.name
            if key in self.functions:
                raise ValueError(
                    f"Function {proto.name!r} from domain "
                    f"{proto.domain!r} as already added."
                )
            self.functions[key] = proto

    def to_onnx(self) -> Union[ModelProto, FunctionProto, GraphProto]:
        """
        Converts the current graph into onnx with the same type
        as the input type.
        """
        if isinstance(self.proto, ModelProto):
            opsets = self.get_opsets()
            inputs = [n.proto for n in self if n.is_input]
            initializer = [n.proto for n in self if n.is_initializer]
            sparse_initializer = [n.proto for n in self if n.is_sparse_initializer]
            nodes = [n.proto for n in self if n.is_node]
            outputs = [n.proto for n in self if n.is_output]
            model = make_model(
                make_graph(
                    nodes,
                    self.proto.graph.name,
                    inputs,
                    outputs,
                    initializer=initializer,
                    sparse_initializer=sparse_initializer,
                ),
                ir_version=self.proto.ir_version,
                producer_name=self.proto.producer_name,
                producer_version=self.proto.producer_version,
                domain=self.proto.domain,
                model_version=self.proto.model_version,
                doc_string=self.proto.doc_string,
                # training_info=self.proto.training_info,
                opset_imports=[make_opsetid(k, v) for k, v in opsets.items()],
                functions=None
                if len(self.functions) == 0
                else list(self.functions.values()),
            )
            if len(self.proto.metadata_props) > 0:
                set_model_props(
                    model, {p.key: p.value for p in self.proto.metadata_props}
                )
            return model

        raise NotImplementedError(
            f"The conversion to onnx is not implemented for type {type(self.proto)}."
        )



from .make_dynamic_quantize_linear import (
    make_dynamic_quantize_linear_function_proto,
    make_simple_dynamic_quantize_linear_function_proto,
)
from .make_reshape_transpose import (
    make_matmul_reshape_transpose_function_proto,
    make_matmul_reshape_transpose_back_function_proto,
)



import os
import sys
from typing import Any, Dict, List, Optional
from argparse import ArgumentParser
from textwrap import dedent


def get_main_parser() -> ArgumentParser:
    parser = ArgumentParser(
        prog="onnx-extended",
        description="onnx-extended main command line.",
        epilog="Type 'python -m onnx_extended <cmd> --help' "
        "to get help for a specific command.",
    )
    parser.add_argument(
        "cmd",
        choices=[
            "store",
            "check",
            "display",
            "print",
            "quantize",
            "select",
            "external",
            "plot",
        ],
        help=dedent(
            """
        Select a command.
        
        'store' executes a model with class CReferenceEvaluator and stores every
        intermediate results on disk with a short onnx to execute the node.
        'check' checks a runtime on stored intermediate results.
        'display' displays the shapes inferences results,
        'print' prints out a model or a protobuf file on the standard output,
        'quantize' quantizes an onnx model in simple ways,
        'select' selects a subgraph inside a bigger models,
        'external' saves the coefficients in a different files for an onnx model,
        'plot' plots a graph like a profiling
        """
        ),
    )
    return parser


def get_parser_store() -> ArgumentParser:
    parser = ArgumentParser(
        prog="store",
        description=dedent(
            """
        Executes a model with class CReferenceEvaluator and stores every
        intermediate results on disk with a short onnx to execute the node.
        """
        ),
        epilog="This is inspired from PR https://github.com/onnx/onnx/pull/5413. "
        "This command may disappear if this functionnality is not used.",
    )
    parser.add_argument(
        "-m",
        "--model",
        type=str,
        required=True,
        help="onnx model to test",
    )
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        default=".",
        help="path where to store the outputs, default is '.'",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=True,
        action="append",
        help="input, it can be a path, or a string like "
        "'float32(4,5)' to generate a random input of this shape, "
        "'rnd' works as well if the model precisely defines the inputs, "
        "'float32(4,5):U10' generates a tensor with a uniform law",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="verbose, default is False",
    )
    parser.add_argument(
        "-r",
        "--runtime",
        choices=["CReferenceEvaluator"],
        default="CReferenceEvaluator",
        help="Runtime to use to generate the intermediate results, "
        "default is 'CReferenceEvaluator'",
    )
    parser.add_argument(
        "-p",
        "--providers",
        default="CPU",
        help="Execution providers, multiple values can separated with a comma",
    )
    return parser


def get_parser_display() -> ArgumentParser:
    parser = ArgumentParser(
        prog="display",
        description=dedent(
            """
        Executes shape inference on an ONNX model and displays the inferred shapes.
        """
        ),
        epilog="This helps looking at a model from a terminal.",
    )
    parser.add_argument(
        "-m",
        "--model",
        type=str,
        required=True,
        help="onnx model to display",
    )
    parser.add_argument(
        "-s",
        "--save",
        type=str,
        required=False,
        help="saved the data as a dataframe",
    )
    parser.add_argument(
        "-t",
        "--tab",
        type=int,
        required=False,
        default=12,
        help="column size when printed on standard output",
    )
    parser.add_argument(
        "-e",
        "--external",
        type=int,
        required=False,
        default=1,
        help="load external data?",
    )
    return parser


def get_parser_print() -> ArgumentParser:
    parser = ArgumentParser(
        prog="print",
        description=dedent(
            """
        Shows an onnx model or a protobuf string on stdout.
        Extension '.onnx' is considered a model,
        extension '.proto' or '.pb' is a protobuf string.
        """
        ),
        epilog="The command can be used on short models, mostly coming "
        "from unittests. Big models are far too large to make this command "
        "useful. Use command display instead.",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=True,
        help="onnx model or protobuf file to print",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["raw", "nodes"],
        default="raw",
        help="format ot use to display the graph",
    )
    parser.add_argument(
        "-e",
        "--external",
        type=int,
        required=False,
        default=1,
        help="load external data?",
    )
    return parser


def get_parser_quantize() -> ArgumentParser:
    parser = ArgumentParser(
        prog="quantize",
        description=dedent(
            """
        Quantizes a model in simple ways.
        """
        ),
        epilog="The implementation quantization are mostly experimental. "
        "Once finalized, the functionality might move to another package.",
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=True,
        help="onnx model or protobuf file to print",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=True,
        help="output model to write",
    )
    parser.add_argument(
        "-k",
        "--kind",
        choices=["fp8", "fp16"],
        required=True,
        help="Kind of quantization to do. 'fp8' "
        "quantizes weights to float 8 e4m3fn whenever possible. "
        "It replaces MatMul by Transpose + DynamicQuantizeLinear + GemmFloat8. "
        "'fp16' casts all float weights to float 16, it does the same for "
        "inputs and outputs. It changes all operators Cast when they "
        "cast into float 32.",
    )
    parser.add_argument(
        "-s",
        "--scenario",
        choices=["onnxruntime", "onnx-extended"],
        required=False,
        default="onnxruntime",
        help="Possible versions for fp8 quantization. "
        "'onnxruntime' uses operators implemented by onnxruntime, "
        "'onnx-extended' uses experimental operators from this package.",
    )
    parser.add_argument(
        "-e",
        "--early-stop",
        required=False,
        help="stops after N modifications",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="do not stop if an exception is raised",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="enable logging, can be repeated",
    )
    parser.add_argument(
        "-t",
        "--transpose",
        default=2,
        help="which input(s) to transpose, 1 for the first one, "
        "2 for the second, 3 for both, 0 for None",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        type=str,
        help="to avoid quantizing nodes if their names belongs "
        "to that list (comma separated)",
    )
    parser.add_argument(
        "-p",
        "--options",
        type=str,
        default="NONE",
        help="options to use for quantization, NONE (default) or OPTIMIZE, "
        "several values can be passed separated by a comma",
    )
    return parser


def get_parser_select() -> ArgumentParser:
    parser = ArgumentParser(
        prog="select",
        description=dedent(
            """
        Selects a subpart of an onnx model.
        """
        ),
        epilog="The function removes the unused nodes.",
    )
    parser.add_argument(
        "-m",
        "--model",
        type=str,
        required=True,
        help="onnx model",
    )
    parser.add_argument(
        "-s",
        "--save",
        type=str,
        required=True,
        help="saves into that file",
    )
    parser.add_argument(
        "-i",
        "--inputs",
        type=str,
        required=False,
        help="list of inputs to keep, comma separated values, "
        "leave empty to keep the model inputs",
    )
    parser.add_argument(
        "-o",
        "--outputs",
        type=str,
        required=False,
        help="list of outputs to keep, comma separated values, "
        "leave empty to keep the model outputs",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="verbose, default is False",
    )
    parser.add_argument(
        "-t",
        "--transpose",
        type=int,
        required=False,
        default=1,
        help="which input to transpose before calling gemm: "
        "0 (none), 1 (first), 2 (second), 3 for both",
    )
    return parser


def get_parser_external() -> ArgumentParser:
    parser = ArgumentParser(
        prog="external",
        description=dedent(
            """
        Takes an onnx model and split the model and the coefficients.    
        """
        ),
        epilog="The functions stores the coefficients as external data. "
        "It calls the function convert_model_to_external_data.",
    )
    parser.add_argument(
        "-m",
        "--model",
        type=str,
        required=True,
        help="onnx model",
    )
    parser.add_argument(
        "-s",
        "--save",
        type=str,
        required=True,
        help="saves into that file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="display sizes",
    )
    return parser


def _cmd_store(argv):
    from ._command_lines import store_intermediate_results

    parser = get_parser_store()
    args = parser.parse_args(argv[1:])
    store_intermediate_results(
        model=args.model,
        runtime=args.runtime,
        verbose=args.verbose,
        inputs=args.input,
        out=args.out,
        providers=args.providers,
    )


def _cmd_display(argv):
    from ._command_lines import display_intermediate_results

    parser = get_parser_display()
    args = parser.parse_args(argv[1:])
    display_intermediate_results(
        model=args.model, save=args.save, tab=args.tab, external=args.external
    )


def _cmd_print(argv):
    from ._command_lines import print_proto

    parser = get_parser_print()
    args = parser.parse_args(argv[1:])
    print_proto(proto=args.input, fmt=args.format, external=args.external)


def _process_exceptions(text: Optional[str]) -> List[Dict[str, str]]:
    if text is None:
        return []
    names = text.split(",")
    return [dict(name=n) for n in names]


def _process_options(text: Optional[str]) -> "QuantizeOptions":  # noqa: F821
    from .tools.graph import QuantizeOptions

    if text is None:
        return QuantizeOptions.NONE
    names = text.split(",")
    value = QuantizeOptions.NONE
    for name in names:
        name = name.upper()
        if hasattr(QuantizeOptions, name):
            value |= getattr(QuantizeOptions, name)
        else:
            raise ValueError(
                f"Unable to parse option name among {dir(QuantizeOptions)}."
            )

    return value


def _cmd_quantize(argv):
    from ._command_lines import cmd_quantize

    parser = get_parser_quantize()
    args = parser.parse_args(argv[1:])
    processed_exceptions = _process_exceptions(args.exclude)
    processed_options = _process_options(args.options)
    cmd_quantize(
        model=args.input,
        output=args.output,
        verbose=args.verbose,
        scenario=args.scenario,
        kind=args.kind,
        early_stop=args.early_stop,
        quiet=args.quiet,
        index_transpose=args.transpose,
        exceptions=processed_exceptions,
        options=processed_options,
    )


def _cmd_select(argv):
    from ._command_lines import cmd_select

    parser = get_parser_select()
    args = parser.parse_args(argv[1:])
    cmd_select(
        model=args.model,
        save=args.save,
        inputs=args.inputs,
        outputs=args.outputs,
        verbose=args.verbose,
    )


def _cmd_external(argv):
    from onnx import load
    from onnx.external_data_helper import (
        convert_model_to_external_data,
        write_external_data_tensors,
    )

    parser = get_parser_external()
    args = parser.parse_args(argv[1:])
    model = args.model
    save = args.save

    if args.verbose:
        size = os.stat(model).st_size
        print(f"Load model {model!r}, size is {size / 2 ** 10:1.3f} kb")
    with open(model, "rb") as f:
        proto = load(f)

    if args.verbose:
        print("convert_model_to_external_data")
    convert_model_to_external_data(
        proto,
        all_tensors_to_one_file=True,
        location=os.path.split(save)[-1] + ".data",
        convert_attribute=True,
        size_threshold=1024,
    )

    dirname = os.path.dirname(save)
    proto = write_external_data_tensors(proto, dirname)
    with open(save, "wb") as f:
        f.write(proto.SerializeToString())
    if args.verbose:
        size = os.stat(save).st_size
        print(f"Saved model {save!r}, size is {size / 2 ** 10:1.3f} kb")


def get_parser_plot() -> ArgumentParser:
    parser = ArgumentParser(
        prog="plot",
        description=dedent(
            """
        Plots a graph representing the data loaded from a
        profiling stored in a filename.
        """
        ),
        epilog="Plots a graph",
    )
    parser.add_argument(
        "-k",
        "--kind",
        choices=[
            "profile_op",
            "profile_node",
        ],
        help=dedent(
            """
        Kind of plot to draw.
        
        'profile_op' shows the time spent in every kernel per operator type,
        'profile_node' shows the time spent in every kernel per operator node,
        """
        ),
        required=True,
    )
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        required=True,
        help="input file",
    )
    parser.add_argument(
        "-c",
        "--ocsv",
        type=str,
        required=False,
        help="saves the data used to plot the graph as csv file",
    )
    parser.add_argument(
        "-o",
        "--opng",
        type=str,
        required=False,
        help="saves the plot as png into that file",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="display sizes",
    )
    parser.add_argument(
        "-w",
        "--with-shape",
        action="store_true",
        help="keep the shape before aggregating the results",
    )
    parser.add_argument(
        "-t",
        "--title",
        required=False,
        help="plot title",
    )
    return parser


def _cmd_plot(argv):
    from ._command_lines import cmd_plot

    parser = get_parser_plot()
    args = parser.parse_args(argv[1:])
    cmd_plot(
        kind=args.kind,
        filename=args.input,
        out_csv=args.ocsv,
        out_png=args.opng,
        title=args.title,
        verbose=args.verbose,
        with_shape=args.with_shape,
    )


def main(argv: Optional[List[Any]] = None):
    if argv is None:
        argv = sys.argv[1:]
    if len(argv) <= 1 or argv[-1] in ("--help", "-h"):
        if len(argv) < 2:
            parser = get_main_parser()
            parser.parse_args(argv)
        else:
            parsers = dict(
                store=get_parser_store,
                print=get_parser_print,
                display=get_parser_display,
                quantize=get_parser_quantize,
                select=get_parser_select,
                external=get_parser_external,
                plot=get_parser_plot,
            )
            cmd = argv[0]
            if cmd not in parsers:
                raise ValueError(
                    f"Unknown command {cmd!r}, it should be in {list(sorted(parsers))}."
                )
            parser = parsers[cmd]()
            parser.parse_args(argv[1:])
        raise RuntimeError("The programme should have exited before.")

    cmd = argv[0]
    fcts = dict(
        store=_cmd_store,
        display=_cmd_display,
        print=_cmd_print,
        quantize=_cmd_quantize,
        select=_cmd_select,
        external=_cmd_external,
        plot=_cmd_plot,
    )
    if cmd in fcts:
        fcts[cmd](argv)
    else:
        raise ValueError(
            f"Unknown command {cmd!r}, use --help to get the list of known command."
        )



import numpy as np
from onnx import TensorProto
from onnx.defs import OpSchema
from onnx.helper import make_attribute
from onnx.reference.custom_element_types import (
    float8e4m3fn,
    float8e4m3fnuz,
    float8e5m2,
    float8e5m2fnuz,
)
from onnx.reference.op_run import OpRun

try:
    from onnx.reference.ops.op_cast import Cast_19 as Cast
    from onnx.reference.ops.op_quantize_linear import (
        QuantizeLinear_19 as QuantizeLinear,
    )
except ImportError:
    from onnx.reference.ops.op_cast import Cast
    from onnx.reference.ops.op_quantize_linear import QuantizeLinear
from onnx.reference.ops.op_dequantize_linear import DequantizeLinear


def _make_schema_gemm_float8(cls_name: str) -> OpSchema:
    """
    Returns the schema for operator `cls_name="GemmFloat8"`.
    """
    gemm_float8_types = [
        "tensor(float8e4m3fn)",
        "tensor(float8e5m2)",
        "tensor(float16)",
        "tensor(bfloat16)",
        "tensor(float)",
    ]
    return OpSchema(
        cls_name,
        "com.microsoft",
        1,
        attributes=[
            OpSchema.Attribute("transA", make_attribute("transA", 0), "transA"),
            OpSchema.Attribute("transB", make_attribute("transB", 0), "transB"),
            OpSchema.Attribute("alpha", make_attribute("alpah", 1.0), "alpha"),
            OpSchema.Attribute("beta", make_attribute("beta", 0.0), "beta"),
            OpSchema.Attribute("smCount", make_attribute("smCount", 0), "smCount"),
            OpSchema.Attribute(
                "fastAccumulationMode",
                make_attribute("fastAccumulationModel", 1),
                "fastAccumulationMode",
            ),
            OpSchema.Attribute(
                "computeType",
                make_attribute("computeType", "CUBLAS_COMPUTE_32F_FAST_TF32"),
                "a value in CUBLAS_COMPUTE_16F, CUBLAS_COMPUTE_32F, "
                "CUBLAS_COMPUTE_32F_FAST_16F, CUBLAS_COMPUTE_32F_FAST_16BF, "
                "CUBLAS_COMPUTE_32F_FAST_TF32",
            ),
            OpSchema.Attribute("rowMajor", make_attribute("rowMajor", 0), "rowMajor"),
            OpSchema.Attribute(
                "dtype", make_attribute("dtype", TensorProto.FLOAT), "dtype"
            ),
            OpSchema.Attribute(
                "activation", OpSchema.AttrType.STRING, "activation", required=None
            ),
        ],
        inputs=[
            OpSchema.FormalParameter("A", "TA"),
            OpSchema.FormalParameter("B", "TB"),
            OpSchema.FormalParameter("C", "TC"),
            OpSchema.FormalParameter("scaleA", "TS"),
            OpSchema.FormalParameter("scaleB", "TS"),
            OpSchema.FormalParameter("scaleY", "TS"),
        ],
        outputs=[OpSchema.FormalParameter("Y", "TR")],
        type_constraints=[
            ("TA", gemm_float8_types, "Constrain input type to input A."),
            ("TB", gemm_float8_types, "Constrain input type to input B."),
            (
                "TC",
                ["tensor(float16)", "tensor(bfloat16)", "tensor(float)"],
                "Constrain input type to input C.",
            ),
            ("TS", ["tensor(float)"], "Constrain input type to input scales."),
            ("TR", gemm_float8_types, "Constrain output type."),
        ],
    )


def is_float8_dtype(dtype) -> bool:
    """
    Returns true if dtype is a float 8 type.
    """
    f8 = {
        (float8e4m3fn, "e4m3fn"),
        (float8e4m3fnuz, "e4m3fnuz"),
        (float8e5m2, "e5m2"),
        (float8e5m2fnuz, "e5m2fnuz"),
    }
    for dt, name in f8:
        if dtype == dt and dtype.descr[0][0] == name:
            return True
    return False


class GemmFloat8(OpRun):
    op_domain = "com.microsoft"
    op_schema = _make_schema_gemm_float8("GemmFloat8")

    def _run(
        self,
        A,
        B,
        *args,
        transA: int = None,
        transB: int = None,
        alpha: float = None,
        beta: float = None,
        smCount: int = None,
        fastAccumulationMode: int = None,
        computeType: str = None,
        dtype: int = None,
        rowMajor: int = None,
        activation: str = None,
    ):
        if len(A.shape) != 2:
            raise ValueError(f"A is not a matrix, its shape is {A.shape}.")
        if len(B.shape) != 2:
            raise ValueError(f"B is not a matrix, its shape is {B.shape}.")
        C = args[0] if len(args) > 0 else None
        scaleA = args[1] if len(args) > 1 else None
        scaleB = args[2] if len(args) > 2 else None
        scaleR = args[3] if len(args) > 3 else None

        if computeType == "CUBLAS_COMPUTE_16F":
            compute_dtype = np.float16
            compute_xtype = TensorProto.FLOAT16
        elif computeType in {
            "CUBLAS_COMPUTE_32F",
            "CUBLAS_COMPUTE_32F_FAST_16F",
            "CUBLAS_COMPUTE_32F_FAST_TF32",
        }:
            compute_dtype = np.float32
            compute_xtype = TensorProto.FLOAT
        else:
            raise ValueError(f"Unexpected value {computeType!r} for computeType.")

        # if rowMajor == 0:
        #     The parameter does not any impact on the result, only on the computation.

        alpha = compute_dtype(alpha)
        beta = compute_dtype(beta)

        fp8a = is_float8_dtype(A.dtype)
        fp8b = is_float8_dtype(B.dtype)
        if fp8a or fp8b:
            if not fp8a or not fp8b:
                raise RuntimeError(
                    f"Both types must be float 8, not one of them only, "
                    f"{A.dtype} @ {B.dtype}."
                )
            if rowMajor == 1:
                raise RuntimeError("Only rowMajor == 0 is supported on GPU.")
            if self.__class__.__name__ == "GemmFloat8" and (transA or not transB):
                raise RuntimeError(
                    f"If both types are float 8, then transA=0 and "
                    f"transB=1 but transA={transA} and transB={transB}."
                )

        ca = Cast.eval(A, to=compute_xtype)
        cb = Cast.eval(B, to=compute_xtype)
        cc = None if C is None else Cast.eval(C, to=compute_xtype)

        if transA:
            ca = ca.T
        if transB:
            cb = cb.T

        if scaleA is not None:
            if scaleA.dtype != np.float32:
                raise ValueError(f"scaleA must be a float not {scaleA.dtype}.")
            ca = DequantizeLinear.eval(ca, scaleA.astype(compute_dtype))
        if scaleB is not None:
            if scaleB.dtype != np.float32:
                raise ValueError(f"scaleB must be a float not {scaleB.dtype}.")
            cb = DequantizeLinear.eval(cb, scaleB.astype(compute_dtype))

        try:
            res = ca @ cb * alpha
        except ValueError as e:
            raise ValueError(
                f"Unable to multiply shapes {ca.shape!r} and {cb.shape!r}."
            ) from e
        if C is not None:
            res += cc * beta

        if scaleR is not None:
            res = QuantizeLinear(res, scaleR.astype(compute_dtype))

        if activation is not None:
            raise NotImplementedError(
                f"activation={activation!r} is not implemented yet."
            )

        final = Cast.eval(res, to=dtype)
        return (final,)


class GemmFloat8Quiet(GemmFloat8):
    op_domain = "com.microsoft"
    op_schema = _make_schema_gemm_float8("GemmFloat8Quiet")



from typing import Any, Dict
import numpy
from onnx import NodeProto
from onnx.defs import OpSchema, get_schema
from onnx.reference.op_run import OpRun
from ..c_ops.cpu.c_op_tree_ensemble_py_ import (
    RuntimeTreeEnsembleRegressorFloat,
    RuntimeTreeEnsembleRegressorDouble,
)


class TreeEnsembleRegressorCommon(OpRun):
    op_domain = "onnx_extented.ortops.optim.cpu"

    def __init__(
        self, onnx_node: NodeProto, run_params: Dict[str, Any], schema: Any = None
    ):
        OpRun.__init__(self, onnx_node, run_params, schema=schema)
        self.parallel = None
        self.rt_ = None
        # default is no parallelization
        self.set_parallel(int(100e6), int(100e6), int(100e6), 1, 1, 0)

    def set_parallel(
        self,
        parallel_tree: int = -1,
        parallel_tree_N: int = -1,
        parallel_N: int = -1,
        batch_size_tree: int = -1,
        batch_size_rows: int = -1,
        node3: int = -1,
    ):
        """
        Sets the parameter for parallelization.
        If a parameter is set to -1, its value does not change.

        :param parallel_tree: parallization by trees if the number of trees is higher
        :param parallel_tree_N: batch size (rows) if parallization by trees
        :param parallel_N: parallization by rows if the number of rows is higher
        :param batch_size_tree: number of trees to compute at the same time
        :param batch_size_rows: number of rows to compute at the same time
        :param node3: use bigger nodes
        """
        self.parallel = (
            parallel_tree,
            parallel_tree_N,
            parallel_N,
            batch_size_tree,
            batch_size_rows,
            node3,
        )
        if self.rt_ is not None:
            self.rt_.set(*self.parallel)

    def _init(self, dtype, **kwargs):
        if dtype == numpy.float32:
            cls = RuntimeTreeEnsembleRegressorFloat
        else:
            cls = RuntimeTreeEnsembleRegressorDouble

        self.rt_ = cls()

        empty_f = numpy.array([], dtype=dtype)
        base_values = numpy.array(
            kwargs.get("base_values", None)
            or kwargs.get("base_values_as_tensor", None)
            or empty_f
        )
        nodes_values = numpy.array(
            kwargs.get("nodes_values", None)
            or kwargs.get("nodes_values_as_tensor", None)
            or empty_f
        )
        nodes_hitrates = numpy.array(
            kwargs.get("nodes_hitrates", None)
            or kwargs.get("nodes_hitrates_as_tensor", None)
            or empty_f
        )
        tw = numpy.array(
            kwargs.get("target_weights", None)
            or kwargs.get("target_weights", None)
            or empty_f
        )

        self.rt_.init(
            kwargs.get("aggregate_function", "SUM"),  # 3
            base_values,  # 4
            kwargs["n_targets"],  # 5
            kwargs["nodes_falsenodeids"],  # 6
            kwargs["nodes_featureids"],  # 7
            nodes_hitrates,  # 8
            kwargs.get("nodes_missing_value_tracks_true", []),  # 9
            kwargs["nodes_modes"].split(","),  # 10
            kwargs["nodes_nodeids"],  # 11
            kwargs["nodes_treeids"],  # 12
            kwargs["nodes_truenodeids"],  # 13
            nodes_values,  # 14
            kwargs["post_transform"],  # 15
            kwargs["target_ids"],  # 16
            kwargs["target_nodeids"],  # 17
            kwargs["target_treeids"],  # 18
            tw,  # 19
        )
        if self.parallel is not None:
            self.rt_.set(*self.parallel)

    def _run(self, x, **kwargs):
        if hasattr(x, "todense"):
            x = x.todense()
        if self.rt_ is None:
            self._init(x.dtype, **kwargs)
        pred = self.rt_.compute(x)
        if pred.shape[0] != x.shape[0]:
            pred = pred.reshape((x.shape[0], -1))
        return (pred,)


def _make_schema():
    attributes = []
    sch = get_schema("TreeEnsembleRegressor", 1, "ai.onnx.ml")
    for att in sch.attributes.values():
        if att.name == "nodes_modes":
            attributes.append(
                OpSchema.Attribute(
                    "nodes_modes",
                    OpSchema.AttrType.STRING,
                    "comma separated value nodes_modes",
                )
            )
        else:
            attributes.append(att)
    return OpSchema(
        "TreeEnsembleRegressor",
        TreeEnsembleRegressorCommon.op_domain,
        1,
        inputs=[
            OpSchema.FormalParameter("X", "T"),
        ],
        outputs=[
            OpSchema.FormalParameter("Y", "T"),
        ],
        type_constraints=[("T", ["tensor(float)"], "")],
        attributes=attributes,
    )


class TreeEnsembleRegressor_1(TreeEnsembleRegressorCommon):
    op_schema = _make_schema()

    def _run(
        self,
        x,
        aggregate_function=None,
        base_values=None,
        n_targets=None,
        nodes_falsenodeids=None,
        nodes_featureids=None,
        nodes_hitrates=None,
        nodes_missing_value_tracks_true=None,
        nodes_modes=None,
        nodes_nodeids=None,
        nodes_treeids=None,
        nodes_truenodeids=None,
        nodes_values=None,
        post_transform=None,
        target_ids=None,
        target_nodeids=None,
        target_treeids=None,
        target_weights=None,
    ):
        return TreeEnsembleRegressorCommon._run(
            self,
            x,
            base_values=base_values,
            n_targets=n_targets,
            nodes_falsenodeids=nodes_falsenodeids,
            nodes_featureids=nodes_featureids,
            nodes_hitrates=nodes_hitrates,
            nodes_missing_value_tracks_true=nodes_missing_value_tracks_true,
            nodes_modes=nodes_modes,
            nodes_nodeids=nodes_nodeids,
            nodes_treeids=nodes_treeids,
            nodes_truenodeids=nodes_truenodeids,
            nodes_values=nodes_values,
            post_transform=post_transform,
            target_ids=target_ids,
            target_nodeids=target_nodeids,
            target_treeids=target_treeids,
            target_weights=target_weights,
        )



# -*- coding: utf-8 -*-
import distutils
import os
import platform
import shutil
import subprocess
import sys
import sysconfig
from pathlib import Path
from typing import List, Tuple

try:
    import numpy
except ImportError as e:
    raise ImportError(
        f"Numpy is not installed, python _executable=f{sys.executable}."
    ) from e

from setuptools import setup, Extension, Command
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.install import install
from wheel.bdist_wheel import bdist_wheel


def get_requirements(here):
    "Returns the requirements from requirements.txt."
    try:
        with open(os.path.join(here, "requirements.txt"), "r") as f:
            requirements = f.read().strip(" \n\r\t").split("\n")
    except FileNotFoundError:
        requirements = []
    if len(requirements) == 0 or requirements == [""]:
        requirements = ["numpy", "scipy", "onnx"]
    return requirements


def get_long_description(here):
    "Returns the long description from README.rst."
    try:
        with open(os.path.join(here, "README.rst"), "r", encoding="utf-8") as f:
            long_description = "onnx-extended:" + f.read().split("onnx-extended:")[1]
    except FileNotFoundError:
        long_description = ""
    return long_description


def get_description():
    return (
        "More operators for onnx reference implementation and "
        "custom kernel implementation for onnxruntime."
    )


def get_version_str(here, default_version):
    VERSION_STR = default_version
    with open(os.path.join(here, "onnx_extended/__init__.py"), "r") as f:
        line = [
            _
            for _ in [_.strip("\r\n ") for _ in f.readlines()]
            if _.startswith("__version__")
        ]
        if len(line) > 0:
            VERSION_STR = line[0].split("=")[1].strip('" ')
    if VERSION_STR is None:
        raise ValueError(f"Unable to guess the package version with here={here!r}.")
    return VERSION_STR


########################################
# C++ Helper
########################################


def find_cuda():
    try:
        p = subprocess.Popen(
            "nvidia-smi",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
    except FileNotFoundError:
        return False
    while True:
        output = p.stdout.readline().decode(errors="ignore")
        if output == "" and p.poll() is not None:
            break
        if output:
            if "CUDA Version:" in output:
                return True
    p.poll()
    return False


def is_windows():
    return platform.system() == "Windows"


def is_darwin():
    return platform.system() == "Darwin"


def _run_subprocess(
    args,
    cwd=None,
    capture_output=False,
    dll_path=None,
    shell=False,
    env=None,
    python_path=None,
    cuda_home=None,
    cuda_version=None,
):
    if env is None:
        env = {}
    if isinstance(args, str):
        raise ValueError("args should be a sequence of strings, not a string")

    my_env = os.environ.copy()
    if cuda_version is not None:
        if is_windows():
            cuda_path = (
                f"C:\\Program Files\\NVIDIA GPU Computing Toolkit\\CUDA\\{cuda_version}"
            )
        elif is_darwin():
            cuda_path = f"/Developer/NVIDIA/CUDA-{cuda_version}"
        else:
            cuda_path = f"/usr/local/cuda-{cuda_version}/bin"
        if "PATH" in my_env:
            my_env["PATH"] = cuda_path + os.pathsep + my_env["PATH"]
        else:
            my_env["PATH"] = cuda_path

    if dll_path:
        if is_windows():
            if "PATH" in my_env:
                my_env["PATH"] = dll_path + os.pathsep + my_env["PATH"]
            else:
                my_env["PATH"] = dll_path
        else:
            if "LD_LIBRARY_PATH" in my_env:
                my_env["LD_LIBRARY_PATH"] += os.pathsep + dll_path
            else:
                my_env["LD_LIBRARY_PATH"] = dll_path

    if is_windows():
        py_path = os.path.dirname(sys.executable)
        if "PATH" in my_env:
            my_env["PATH"] = py_path + os.pathsep + my_env["PATH"]
        else:
            my_env["PATH"] = py_path

    # Add nvcc's folder to PATH env so that our cmake file can find nvcc
    if cuda_home:
        my_env["PATH"] = os.path.join(cuda_home, "bin") + os.pathsep + my_env["PATH"]

    if python_path:
        if "PYTHONPATH" in my_env:
            my_env["PYTHONPATH"] += os.pathsep + python_path
        else:
            my_env["PYTHONPATH"] = python_path

    my_env.update(env)

    p = subprocess.Popen(
        args,
        cwd=cwd,
        shell=shell,
        env=my_env,
        stdout=subprocess.PIPE if capture_output else None,
        stderr=subprocess.STDOUT if capture_output else None,
    )
    raise_exception = False
    while True:
        output = p.stdout.readline().decode(errors="ignore")
        if output == "" and p.poll() is not None:
            break
        if output:
            out = output.rstrip()
            sys.stdout.write(out + "\n")
            sys.stdout.flush()
            if (
                "fatal error" in output
                or "CMake Error" in output
                or "gmake: ***" in output
                or "): error C" in output
                or ": error: " in output
            ):
                raise_exception = True
    rc = p.poll()
    if raise_exception:
        raise RuntimeError("An error was found in the output. The build is stopped.")
    return rc


########################################
# C++ CMake Extension
########################################


class CMakeExtension(Extension):
    def __init__(self, name: str, library: str = "") -> None:
        super().__init__(name, sources=[])
        print(f"-- setup: add extension {name}")
        self.library_file = os.fspath(Path(library).resolve())


class cmake_build_class_extension(Command):
    user_options = [
        *build_ext.user_options,
        (
            "use-cuda=",
            None,
            "If cuda is available, CUDA is "
            "used by default unless this option is set to 0",
        ),
        ("use-nvtx=", None, "Enables compilation with NVTX events."),
        (
            "cuda-version=",
            None,
            "If cuda is available, it searches the installed version "
            "unless this option is defined.",
        ),
        (
            "parallel=",
            None,
            "Parallelization",
        ),
        (
            "ort-version=",
            None,
            "onnxruntime version, a path is allowed",
        ),
        (
            "cuda-build=",
            None,
            "CUDA code can be compiled to be working with "
            "different architectures, this flag can optimize "
            "for a specific machine, possible values: DEFAULT, "
            "H100, H100opt",
        ),
        (
            "cuda-link=",
            None,
            "CUDA can statically linked (STATIC) or dynamically "
            "(SHARED), default is STATIC."
            "STATIC",
        ),
        (
            "manylinux=",
            None,
            "Enforces the compilation with manylinux, " "default is set to 0.",
        ),
    ]

    def initialize_options(self):
        self.use_nvtx = None
        self.use_cuda = None
        self.cuda_version = None
        self.parallel = None
        self.manylinux = None
        self.ort_version = DEFAULT_ORT_VERSION
        self.cuda_build = "DEFAULT"
        self.cuda_link = "STATIC"

        self._parent.initialize_options(self)

        # boolean
        b_values = {0, 1, "1", "0", True, False}
        t_values = {1, "1", True}
        for att in ["use_nvtx", "use_cuda", "manylinux"]:
            v = getattr(self, att)
            if v is not None:
                continue
            v = os.environ.get(att.upper(), None)
            if v is None:
                continue
            if v not in b_values:
                raise ValueError(f"Unable to interpret value {v} for {att.upper()!r}.")
            print(f"-- setup: use env {att.upper()}={v in t_values}")
            setattr(self, att, v in t_values)

        if self.ort_version is None:
            self.ort_version = os.environ.get("ORT_VERSION", None)
            if self.ort_version not in ("", None):
                print(f"-- setup: use env ORT_VERSION={self.ort_version}")
        if self.cuda_build is None:
            self.cuda_build = os.environ.get("CUDA_BUILD", None)
            if self.cuda_build not in ("", None):
                print(f"-- setup: use env CUDA_BUILD={self.cuda_build}")
        if self.cuda_version is None:
            self.cuda_version = os.environ.get("CUDA_VERSION", None)
            if self.cuda_version not in ("", None):
                print(f"-- setup: use env CUDA_VERSION={self.cuda_version}")
        if self.use_nvtx is None:
            self.use_nvtx = False
        if self.manylinux is None:
            self.manylinux = False

    def finalize_options(self):
        self._parent.finalize_options(self)

        b_values = {0, 1, "1", "0", True, False, "True", "False"}
        if self.use_nvtx not in b_values:
            raise ValueError(f"use_nvtx={self.use_nvtx!r} must be in {b_values}.")
        if self.use_cuda is None:
            self.use_cuda = find_cuda()
        if self.use_cuda not in b_values:
            raise ValueError(f"use_cuda={self.use_cuda!r} must be in {b_values}.")

        self.use_nvtx = self.use_nvtx in {1, "1", True, "True"}
        self.use_cuda = self.use_cuda in {1, "1", True, "True"}
        self.manylinux = self.manylinux in {1, "1", True, "True"}

        if self.cuda_version in (None, ""):
            self.cuda_version = None
        build = {"DEFAULT", "H100", "H100opt"}
        if self.cuda_build not in build:
            raise ValueError(f"cuda-build={self.cuda_build!r} not in {build}.")
        link = {"STATIC", "SHARED"}
        if self.cuda_link not in link:
            raise ValueError(f"cuda-link={self.cuda_link!r} not in {link}.")

        options = {o[0]: o for o in self.user_options}
        keys = list(sorted(options.keys()))
        for na in keys:
            opt = options[na]
            name = opt[0].replace("-", "_").strip("=")
            print(f"-- setup: option {name}={getattr(self, name, None)}")

    def get_cmake_args(self, cfg: str) -> List[str]:
        """
        Returns the argument for cmake.

        :param cfg: configuration (Release, ...)
        :return: build_path, self.build_lib
        """
        iswin = is_windows()
        isdar = is_darwin()
        cmake_cmd_args = []

        path = sys.executable
        vers = (
            f"{sys.version_info.major}."
            f"{sys.version_info.minor}."
            f"{sys.version_info.micro}"
        )
        versmm = f"{sys.version_info.major}.{sys.version_info.minor}"
        module_ext = distutils.sysconfig.get_config_var("EXT_SUFFIX")

        here = os.path.dirname(__file__)
        if here == "":
            here = "."

        manylinux_tags = [
            "manylinux1_x86_64",
            "manylinux1_i686",
            "manylinux2010_x86_64",
            "manylinux2010_i686",
            "manylinux2014_x86_64",
            "manylinux2014_i686",
            "manylinux2014_aarch64",
            "manylinux2014_armv7l",
            "manylinux2014_ppc64",
            "manylinux2014_ppc64le",
            "manylinux2014_s390x",
            "manylinux_2_24_x86_64",
            "manylinux_2_24_aarch64",
            "manylinux_2_28_x86_64",
            "manylinux_2_28_aarch64",
            "manylinux_2_35_x86_64",
            "manylinux_2_35_aarch64",
        ]
        is_manylinux = (
            self.manylinux or os.environ.get("AUDITWHEEL_PLAT", None) in manylinux_tags
        )

        cmake_args = [
            f"-DPYTHON_EXECUTABLE={path}",
            f"-DCMAKE_BUILD_TYPE={cfg}",
            f"-DPYTHON_VERSION={vers}",
            f"-DPYTHON_VERSION_MM={versmm}",
            f"-DPYTHON_MODULE_EXTENSION={module_ext}",
            f"-DORT_VERSION={self.ort_version}",
            f"-DONNX_EXTENDED_VERSION={get_version_str(here, None)}",
            f"-DPYTHON_MANYLINUX={1 if is_manylinux else 0}",
        ]
        if self.parallel is not None:
            cmake_args.append(f"-j{self.parallel}")

        if self.use_nvtx:
            cmake_args.append("-DUSE_NVTX=1")
        cmake_args.append(f"-DUSE_CUDA={1 if self.use_cuda else 0}")
        if self.use_cuda:
            cmake_args.append(f"-DCUDA_BUILD={self.cuda_build}")
            cmake_args.append(f"-DCUDA_LINK={self.cuda_link}")
        cuda_version = self.cuda_version
        if cuda_version not in (None, ""):
            cmake_args.append(f"-DCUDA_VERSION={cuda_version}")

        if iswin or isdar:
            include_dir = sysconfig.get_paths()["include"].replace("\\", "/")
            lib_dir = (
                sysconfig.get_config_var("LIBDIR")
                or sysconfig.get_paths()["stdlib"]
                or ""
            ).replace("\\", "/")
            numpy_include_dir = numpy.get_include().replace("\\", "/")
            cmake_args.extend(
                [
                    f"-DPYTHON_INCLUDE_DIR={include_dir}",
                    # f"-DPYTHON_LIBRARIES={lib_dir}",
                    f"-DPYTHON_LIBRARY_DIR={lib_dir}",
                    f"-DPYTHON_NUMPY_INCLUDE_DIR={numpy_include_dir}",
                    # "-DUSE_SETUP_PYTHON=1",
                    f"-DPYTHON_NUMPY_VERSION={numpy.__version__}",
                ]
            )
            os.environ["PYTHON_NUMPY_INCLUDE_DIR"] = numpy_include_dir

        cmake_args += cmake_cmd_args
        return cmake_args

    def build_cmake(self, cfg: str, cmake_args: List[str]) -> Tuple[str, str]:
        """
        Calls cmake.

        :param cfg: configuration (Release, ...)
        :param cmake_args: cmake aguments
        :return: build_path, self.build_lib
        """
        this_dir = os.path.dirname(os.path.abspath(__file__))
        build_temp = getattr(self, "build_temp", None)
        if build_temp is None:
            build_temp = os.path.join(this_dir, "build", "install")

        if not os.path.exists(build_temp):
            os.makedirs(build_temp)

        # Builds the project.
        build_path = os.path.abspath(build_temp)
        build_lib = getattr(self, "build_lib", build_path)
        cmake_args = cmake_args + [
            f"-DSETUP_BUILD_PATH={os.path.abspath(build_path)}",
            f"-DSETUP_BUILD_LIB={os.path.abspath(build_lib)}",
        ]
        with open(
            os.path.join(os.path.dirname(__file__), ".build_path.txt"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(build_path)
        # build_path = os.path.join(this_dir, "build")
        if not os.path.exists(build_path):
            os.makedirs(build_path)
        source_path = os.path.join(this_dir, "_cmake")

        cmd = ["cmake", "-S", source_path, "-B", build_path, *cmake_args]
        print(f"-- setup: version={sys.version_info!r}")
        print(f"-- setup: cwd={os.getcwd()!r}")
        print(f"-- setup: source_path={source_path!r}")
        print(f"-- setup: build_path={build_path!r}")
        print(f"-- setup: cmd={' '.join(cmd)}")
        _run_subprocess(
            cmd, cwd=build_path, capture_output=True, cuda_version=self.cuda_version
        )

        # then build
        print()
        cmd = ["cmake", "--build", build_path, "--config", cfg]
        print(f"-- setup: cwd={os.getcwd()!r}")
        print(f"-- setup: build_path={build_path!r}")
        print(f"-- setup: cmd={' '.join(cmd)}")
        _run_subprocess(
            cmd, cwd=build_path, capture_output=True, cuda_version=self.cuda_version
        )
        print("-- setup: done.")
        return build_path, build_lib

    def process_extensions(self, cfg: str, build_path: str, build_lib: str):
        """
        Copies the python extensions built by cmake into python subfolders.

        :param cfg: configuration (Release, ...)
        :param build_path: where it was built
        :param build_lib: built library
        """
        if not hasattr(self, "extensions"):
            raise RuntimeError(f"Unable to get the list of extensions: {dir(self)}.")
        iswin = is_windows()
        for ext in self.extensions:
            full_name = ext._file_name
            name = os.path.split(full_name)[-1]
            if iswin:
                looks = [
                    os.path.join(build_path, cfg, full_name),
                    os.path.join(build_path, cfg, name),
                ]
            else:
                looks = [
                    os.path.join(build_path, full_name),
                    os.path.join(build_path, name),
                ]
            looks_exists = [look for look in looks if os.path.exists(look)]
            if len(looks_exists) == 0:
                raise FileNotFoundError(
                    f"Unable to find {name!r} as {looks!r} (full_name={full_name!r}), "
                    f"build_path contains {os.listdir(build_path)}."
                )
            else:
                look = looks_exists[0]
            dest = os.path.join(build_lib, os.path.split(full_name)[0])
            if not os.path.exists(dest):
                os.makedirs(dest)
            if not os.path.exists(look):
                raise FileNotFoundError(f"Unable to find {look!r}.")
            if not os.path.exists(dest):
                raise FileNotFoundError(f"Unable to find folder {dest!r}.")
            print(f"-- setup: copy-2 {look!r} to {dest!r}")
            shutil.copy(look, dest)

    def run_cmake(self):
        # Ensure that CMake is present and working
        try:
            subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError("Cannot find CMake executable")

        cfg = "Release"
        cmake_args = self.get_cmake_args(cfg)
        build_path, build_lib = self.build_cmake(cfg, cmake_args)
        if hasattr(self, "extensions"):
            print("-- process_extensions")
            self.process_extensions(cfg, build_path, build_lib)
        else:
            print("-- skip process_extensions")
        print("-- done")


class cmake_build_ext(cmake_build_class_extension, build_ext):
    _parent = build_ext
    user_options = build_ext.user_options + cmake_build_class_extension.user_options

    def run(self):
        # cmake_build_class_extension.run_cmake(self)
        return build_ext.run(self)

    def build_extensions(self):
        self.run_cmake()


class cmake_build_py(cmake_build_class_extension, build_py):
    _parent = build_py
    user_options = build_py.user_options + cmake_build_class_extension.user_options

    def run(self):
        return build_py.run(self)


class cmake_develop(cmake_build_class_extension, develop):
    _parent = develop
    user_options = develop.user_options + cmake_build_class_extension.user_options

    def run(self):
        return develop.run(self)


class cmake_install(cmake_build_class_extension, install):
    _parent = install
    user_options = install.user_options + cmake_build_class_extension.user_options

    def run(self):
        return install.run(self)


class cmake_bdist_wheel(cmake_build_class_extension, bdist_wheel):
    _parent = bdist_wheel
    user_options = bdist_wheel.user_options + cmake_build_class_extension.user_options

    def run(self):
        return bdist_wheel.run(self)


def get_ext_modules():
    if is_windows():
        ext = "pyd"
    elif is_darwin():
        ext = "dylib"
    else:
        ext = "so"

    cuda_extensions = []
    has_cuda = find_cuda()
    if has_cuda:
        add_cuda = True
        if "--use-cuda" in sys.argv:
            pos = sys.argv.index("--use-cuda")
            if len(sys.argv) > pos + 1 and sys.argv[pos + 1] in (
                "0",
                0,
                False,
                "False",
            ):
                add_cuda = False
        elif "--use-cuda=0" in sys.argv:
            add_cuda = False
        elif os.environ.get("USE_CUDA", None) in {0, "0", False}:
            add_cuda = False
        if add_cuda:
            cuda_extensions.extend(
                [
                    CMakeExtension(
                        "onnx_extended.validation.cuda.cuda_example_py",
                        f"onnx_extended/validation/cuda/cuda_example_py.{ext}",
                    )
                ]
            )
    elif "--use-cuda=1" in sys.argv or "--use-cuda" in sys.argv:
        raise RuntimeError(
            "CUDA is not available, it cannot be build with CUDA depsite "
            "option '--use-cuda=1'."
        )
    ext_modules = [
        CMakeExtension(
            "onnx_extended.validation.cython.vector_function_cy",
            f"onnx_extended/validation/cython/vector_function_cy.{ext}",
        ),
        CMakeExtension(
            "onnx_extended.validation.cython.fp8",
            f"onnx_extended/validation/cython/fp8.{ext}",
        ),
        CMakeExtension(
            "onnx_extended.validation.cpu._validation",
            f"onnx_extended/validation/cpu/_validation.{ext}",
        ),
        CMakeExtension(
            "onnx_extended.reference.c_ops.cpu.c_op_conv_",
            f"onnx_extended/reference/c_ops/cpu/c_op_conv_.{ext}",
        ),
        CMakeExtension(
            "onnx_extended.reference.c_ops.cpu.c_op_tree_ensemble_py_",
            f"onnx_extended/reference/c_ops/cpu/c_op_tree_ensemble_py_.{ext}",
        ),
        CMakeExtension(
            "onnx_extended.ortcy.wrap.ortinf",
            f"onnx_extended.ortcy.wrap.ortinf.{ext}",
        ),
        *cuda_extensions,
    ]
    return ext_modules


######################
# beginning of setup
######################

DEFAULT_ORT_VERSION = "1.16.0"
here = os.path.dirname(__file__)
if here == "":
    here = "."


def get_package_data():
    known_extensions = [
        "*.cc",
        "*.cpp",
        "*.cu",
        "*.cuh",
        "*.dylib",
        "*.h",
        "*.hpp",
        "*.pyd",
        "*.pyx",
        "*.so*",
        "*.dll",
    ]
    return {
        "onnx_extended": known_extensions,
        "onnx_extended.cpp": known_extensions,
        "onnx_extended.include": known_extensions,
        "onnx_extended.include.common": known_extensions,
        "onnx_extended.include.cpu": known_extensions,
        "onnx_extended.include.cuda": known_extensions,
        "onnx_extended.ortops.optim.cpu": known_extensions,
        "onnx_extended.ortops.tutorial.cpu": known_extensions,
        "onnx_extended.ortops.tutorial.cuda": known_extensions,
        "onnx_extended.ortcy.wrap": known_extensions,
        "onnx_extended.reference.c_ops.cpu": known_extensions,
        "onnx_extended.validation.cpu": known_extensions,
        "onnx_extended.validation.cython": known_extensions,
        "onnx_extended.validation.cuda": known_extensions,
    }


setup(
    name="onnx-extended",
    version=get_version_str(here, "0.2.2"),
    description=get_description(),
    long_description=get_long_description(here),
    author="Xavier Dupré",
    author_email="xavier.dupre@gmail.com",
    url="https://github.com/sdpython/onnx-extended",
    package_data=get_package_data(),
    setup_requires=["numpy", "scipy", "onnx"],
    install_requires=get_requirements(here),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: C",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    cmdclass={
        "build_ext": cmake_build_ext,
        "build_py": cmake_build_py,
        "bdist_wheel": cmake_bdist_wheel,
        "cmake_build": cmake_build_class_extension,
        "develop": cmake_develop,
        "install": cmake_install,
    },
    ext_modules=get_ext_modules(),
)



import os
import re
import io
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import numpy as np
from onnx import (
    FunctionProto,
    GraphProto,
    ModelProto,
    TensorProto,
    SequenceProto,
    ValueInfoProto,
    load,
)
from onnx.helper import tensor_dtype_to_np_dtype
from onnx.numpy_helper import to_array
from onnx.reference import ReferenceEvaluator


def _type_shape(
    input_def: Union[str, ValueInfoProto]
) -> Tuple[Any, Tuple[Union[int, str], ...], Any]:
    if isinstance(input_def, str):
        reg = re.compile(
            "([a-z][a-z0-9]*)?([(]([ a-zA-Z,0-9]+)[)])?(:([A-Z][A-Z0-9]*))?"
        )
        search = reg.match(input_def)
        if search is None:
            raise ValueError(f"Unable to interpret string {input_def!r}.")
        grs = search.groups()
        dtype = grs[0]
        shape = None if grs[2] is None else grs[2].replace(" ", "").split(",")
        law = grs[-1]
        new_shape = []
        if shape is not None:
            for i in shape:
                try:
                    vi = int(i)
                    new_shape.append(vi)
                except ValueError:
                    new_shape.append(i)
            rshape = tuple(new_shape)
        else:
            rshape = None
        dt = None if dtype is None else getattr(np, dtype)
        return dt, rshape, law

    if isinstance(input_def, ValueInfoProto):
        try:
            ttype = input_def.type.tensor_type
        except AttributeError:
            raise ValueError(f"Unsupported input type {input_def!r}.")
        dt = ttype.elem_type
        new_shape = []
        for d in ttype.shape.dim:
            if d.dim_param:
                new_shape.append(d.dim_param)
            else:
                new_shape.append(d.dim_value)
        ndt = tensor_dtype_to_np_dtype(dt)
        return ndt, tuple(new_shape), None

    raise TypeError(f"Unexpected type {type(input_def)} for input_def.")


def _generate_random_inputs(
    dtype: Any,
    shape: Tuple[Union[int, str], ...],
    law: Optional[str] = None,
    dims: Optional[Dict[str, int]] = None,
) -> Tuple[np.ndarray, Dict[str, int]]:
    """
    Creates random or specific inputs.

    :param dtype: numpy dtype
    :param shape: expected shape
    :param law: law of the coefficients, default is 'U10', uniform law
    :param dims: letter are allowed, contains the named dimensions already
        mapped to a specific value
    :return: tuple (array, updated dims)

    Dimensions are modified inplace.
    """
    if dims is None:
        dims = {}
    if law is None:
        law = "U10"
    new_shape = []
    for sh in shape:
        if isinstance(sh, int):
            new_shape.append(sh)
        elif isinstance(sh, str):
            if sh not in dims:
                dims[sh] = 8
            new_shape.append(dims[sh])
    final_shape = tuple(new_shape)

    if law == "U10":
        res = np.random.random(final_shape).astype(dtype)
        return res, dims

    raise ValueError(f"Unexpected value for law={law!r}.")


def store_intermediate_results(
    model: Union[ModelProto, str],
    inputs: List[Union[str, np.ndarray, TensorProto]],
    out: str = ".",
    runtime: Union[type, str] = "CReferenceEvaluator",
    providers: Union[str, List[str]] = "CPU",
    verbose: int = 0,
):
    """
    Executes an onnx model with a runtime and stores the
    intermediate results in a folder.
    See :class:`CReferenceEvaluator <onnx_extended.reference.CReferenceEvaluator>`
    for further details.

    :param model: path to a model or ModelProto
    :param inputs: list of inputs for the model
    :param out: output path
    :param runtime: runtime class to use
    :param providers: list of providers
    :param verbose: verbosity level
    :return: outputs
    """
    if isinstance(model, str):
        if not os.path.exists(model):
            raise FileNotFoundError(f"File {model!r} does not exists.")
    if isinstance(providers, str):
        providers = [s.strip() for s in providers.split(",")]
    if runtime == "CReferenceEvaluator":
        from .reference import CReferenceEvaluator

        cls_runtime = CReferenceEvaluator
        add_providers = False
    elif hasattr(runtime, "run"):
        cls_runtime = runtime
        add_providers = True
    else:
        raise ValueError(f"Unexpected runtime {runtime!r}.")
    iv = int(verbose)
    if add_providers:
        inst = cls_runtime(
            model, providers=providers, save_intermediate=out, verbose=iv
        )
    else:
        inst = cls_runtime(model, save_intermediate=out, verbose=iv)
    names = inst.input_names
    if len(names) < len(inputs):
        raise RuntimeError(
            f"There are more inputs ({len(inputs)}) "
            f"than names ({len(names)}). Names are {names}."
        )

    dims = {}
    feeds = {}
    for i, (name, inp) in enumerate(zip(names, inputs)):
        if isinstance(inp, str) and os.path.exists(inp):
            with open(inp, "rb") as f:
                vect = f.read()
            tp = TensorProto()
            tp.ParseFromString(vect)
            value = to_array(tp)
        else:
            ty, shape, law = _type_shape(inp)
            if ty is None or shape is None:
                if isinstance(inst, ReferenceEvaluator):
                    ty, shape, _ = _type_shape(inst.proto_.graph.input[i])
                else:
                    raise RuntimeError(
                        f"shape or dtype is unknown and cannot "
                        f"be retrieved from class {type(inst)}."
                    )
            value, dims = _generate_random_inputs(ty, shape, law, dims=dims)
        feeds[name] = value

    got = inst.run(None, feeds)
    return got


def display_intermediate_results(
    model: str,
    save: Optional[str] = None,
    tab: int = 12,
    external: bool = True,
    fprint: Callable = print,
):
    """
    Displays shape, type for a model.

    :param model: a model
    :param save: save the results as a dataframe
    :param tab: column size for the output
    :param external: loads the external data or not
    :param fprint: function to print
    """
    from .tools.onnx_tools import enumerate_onnx_node_types

    if save is not None:
        ext = os.path.splitext(save)[-1]
        if ext not in {".csv", ".xlsx"}:
            raise ValueError(f"Unexpected format {save!r}, extension is {ext!r}.")
    else:
        exts = None

    def _fixed(s, length=10):
        if not isinstance(s, str):
            raise TypeError(f"Unexpected type {type(s)}: {s!r}.")
        return (
            (s[: length - 1] + " ")
            if len(s) >= length - 1
            else s + " " * (length - len(s))
        )

    n_rows = 0
    rows = []
    for obs in enumerate_onnx_node_types(model, external=external):
        if "level" not in obs:
            raise RuntimeError(f"Unexpected value obs={obs!r}.")
        indent = " " * obs["level"] * tab
        values = [
            indent,
            _fixed(obs.get("kind", ""), tab),
            _fixed(obs.get("type", ""), tab),
            _fixed(obs.get("name", ""), tab),
            _fixed(obs.get("elem_type", ""), tab),
            _fixed(obs.get("shape", ""), tab),
            _fixed(obs.get("input_types", ""), tab),
            _fixed(obs.get("output_types", ""), tab),
            _fixed(obs.get("inputs", ""), tab),
            _fixed(obs.get("outputs", ""), tab),
        ]
        line = "".join(values)
        fprint(line)

        n_rows += 1
        if save is not None:
            rows.append(obs)

    if n_rows == 0:
        if isinstance(model, str):
            raise RuntimeError(f"Model {model!r} is empty.")
        raise RuntimeError(f"Model type {type(model)} is empty.")

    if save is not None:
        from pandas import DataFrame

        df = DataFrame(rows)
        exts = {".csv": df.to_csv, ".xlsx": df.to_excel}
        exts[ext](save, index=False)


def print_proto(proto: str, fmt: str = "raw", external: bool = True):
    """
    Shows an onnx model or a protobuf string on stdout.
    Extension '.onnx' is considered a model,
    extension '.proto' or '.pb' is a protobuf string.

    :param proto: a file
    :param fmt: format to use to print the model,
        `raw` prints out the string produced by `print(model)`,
        `nodes` only prints out the node name
    :param external: loads with external data
    """
    if isinstance(proto, str):
        if not os.path.exists(proto):
            raise FileNotFoundError(f"Unable to find file {proto!r}.")
        ext = os.path.splitext(proto)[-1]
        if ext == ".onnx":
            with open(proto, "rb") as f:
                proto_loaded = load(f, load_external_data=external)
        elif ext in (".pb", ".proto"):
            with open(proto, "rb") as f:
                content = f.read()
            exc = []
            proto_loaded = None
            for cls in [
                TensorProto,
                SequenceProto,
                FunctionProto,
                ModelProto,
                GraphProto,
            ]:
                inst = cls()
                try:
                    inst.ParseFromString(content)
                    proto_loaded = inst
                    break
                except Exception as e:
                    exc.append((cls, e))
            if proto_loaded is None:
                msg = "\n".join(f"type: {c}: {e}" for c, e in exc)
                raise RuntimeError(f"Unable to load {proto!r}, tried:\n{msg}")
        else:
            raise ValueError(f"Unexpected file extension {ext!r} for file {proto!r}.")
    else:
        proto_loaded = proto

    print(f"Type: {type(proto_loaded)}")
    if fmt == "raw":
        print(proto_loaded)
    elif fmt == "nodes":
        from .tools.graph.onnx_graph_struct import Graph

        if proto_loaded is None:
            raise ValueError(f"Unable to load {proto!r}.")
        graph = Graph(proto_loaded)
        for node in graph:
            print(str(node).replace("<parent>, ", ""))
    else:
        raise ValueError(f"Unexpected value for fmt={fmt!r}.")


def cmd_quantize(
    model: Union[ModelProto, str],
    output: Optional[str] = None,
    kind: str = "fp8",
    scenario: str = "onnxruntime",
    early_stop: Optional[int] = None,
    quiet: bool = False,
    verbose: int = 0,
    index_transpose: int = 2,
    exceptions: Optional[List[Dict[str, str]]] = None,
    options: Optional["QuantizeOptions"] = None,  # noqa: F821
):
    """
    Quantizes a model

    :param model: path to a model or ModelProto
    :param output: output file
    :param kind: kind of quantization
    :param scenario: depends on the quantization
    :param early_stop: stops early to see the preliminary results
    :param quiet: do not stop an exception
    :param verbose: verbosity level
    :param index_transpose: which input to transpose before calling gemm:
        0 (none), 1 (first), 2 (second), 3 for both
    :param exceptions: exclude nodes from the quantization,
        `[{"name": "node_name1"}, {"name": "node_name2"}]` will exclude
        these two node names from the quantization
    :param options: quantization options, see class
        :class:`QuantizeOptions <onnx_extended.tools.graph.QuantizeOptions>`
    """
    from .tools.graph import Graph, QuantizeOptions

    if options is None:
        options = QuantizeOptions.NONE

    if isinstance(model, str):
        if not os.path.exists(model):
            raise FileNotFoundError(f"Unable to find file {model!r}.")
        ext = os.path.splitext(model)[-1]
        if ext == ".onnx":
            with open(model, "rb") as f:
                proto_loaded = load(f)
    else:
        proto_loaded = model
    graph = Graph(proto_loaded)

    if verbose:
        logging.basicConfig(
            level=logging.WARN
            if verbose > 2
            else (logging.DEBUG if verbose > 1 else logging.INFO)
        )

    if kind == "fp8":
        from .tools.graph import quantize_float8

        logger = logging.getLogger("onnx-extended")
        logger.info("Model initial size: %d", len(proto_loaded.SerializeToString()))
        new_graph = quantize_float8(
            graph,
            early_stop=early_stop or -1,
            quiet=quiet,
            version=scenario,
            index_transpose=index_transpose,
            exceptions=exceptions,
            quantize_options=options,
        )
        if new_graph is None:
            logger.warning("No node was quantized.")
            return
        onx2 = new_graph.to_onnx()
        seq = onx2.SerializeToString()
        logger.info("Model quantized size: %d", len(seq))
        if output is not None:
            with open(output, "wb") as f:
                f.write(seq)
        return

    if kind == "fp16":
        from .tools.graph import cast_constant

        logger = logging.getLogger("onnx-extended")
        logger.info("Model initial size: %d", len(proto_loaded.SerializeToString()))
        new_graph = cast_constant(
            graph,
            quiet=quiet,
            from_type=TensorProto.FLOAT,
            to_type=TensorProto.FLOAT16,
        )
        if new_graph is None:
            logger.warning("No node was modified.")
            return
        onx2 = new_graph.to_onnx()
        seq = onx2.SerializeToString()
        logger.info("Model reduced size: %d", len(seq))
        if output is not None:
            with open(output, "wb") as f:
                f.write(seq)
        return

    raise ValueError(f"Unexpected value {kind!r} for kind.")


def cmd_select(
    model: Union[ModelProto, str],
    save: Optional[str] = None,
    inputs: Optional[Union[str, List[str]]] = None,
    outputs: Optional[Union[str, List[str]]] = None,
    verbose: int = 0,
):
    """
    Selects a subgraph in a model.

    :param model: path to a model or ModelProto
    :param save: model ot save in this file
    :param inputs: list of inputs or empty to keep the original inputs
    :param outputs: list of outputs or empty to keep the original outputs
    :param verbose: verbosity level
    """
    from .tools.onnx_manipulations import select_model_inputs_outputs

    if isinstance(model, str):
        if not os.path.exists(model):
            raise FileNotFoundError(f"Unable to find file {model!r}.")
        ext = os.path.splitext(model)[-1]
        if ext == ".onnx":
            with open(model, "rb") as f:
                proto_loaded = load(f)
    else:
        proto_loaded = model

    if verbose:
        logging.basicConfig(
            level=logging.WARN
            if verbose > 2
            else (logging.DEBUG if verbose > 1 else logging.INFO)
        )

    if isinstance(inputs, str):
        inputs = inputs.strip().split(",")
    if isinstance(outputs, str):
        outputs = outputs.strip().split(",")

    logger = logging.getLogger("onnx-extended")
    logger.info("Initial model size: %d", len(proto_loaded.SerializeToString()))
    onx2 = select_model_inputs_outputs(
        proto_loaded,
        inputs=inputs,
        outputs=outputs,
        verbose=verbose,
    )
    seq = onx2.SerializeToString()
    logger.info("Selected model size: %d", len(seq))
    if save is not None:
        with open(save, "wb") as f:
            f.write(seq)


def plot_profile(
    filename: str,
    kind: str,
    out_csv: Optional[str] = None,
    out_png: Optional[str] = None,
    title: Optional[str] = None,
    with_shape: bool = False,
    verbose: int = 0,
):
    """
    Plots a profiling.

    :param filename: raw data to load
    :param kind: kind of plot to so, see below
    :param out_csv: output the data into that csv file
    :param out_png: output the graph in that file
    :param with_shape: consider input shape when showing results
    :param title: title (optional)
    :param verbose: verbosity, if > 0, prints out the data in csv format
    """
    import matplotlib.pyplot as plt
    from .tools.js_profile import (
        js_profile_to_dataframe,
        plot_ort_profile,
        _preprocess_graph1,
        _preprocess_graph2,
    )

    if verbose:
        print(f"[plot_profile] load {filename!r}")

    if kind == "profile_op":
        df = js_profile_to_dataframe(filename, first_it_out=True, with_shape=with_shape)

        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        plot_ort_profile(df, ax, title=title)
        df = _preprocess_graph1(df)
    elif kind == "profile_node":
        df = js_profile_to_dataframe(
            filename, first_it_out=True, agg=True, with_shape=with_shape
        )

        fig, ax = plt.subplots(1, 1, figsize=(5, 5))
        plot_ort_profile(df, ax, title=title)
        df = _preprocess_graph2(df)
    else:
        raise ValueError(f"Unexpected kind {kind:r}.")

    if verbose:
        st = io.StringIO()
        df.to_csv(st)
        print(st.getvalue())

    if out_csv not in {"", None}:
        if verbose:
            print(f"[plot_profile] save {out_csv!r}")
        df.to_csv(out_csv)

    if out_png not in {"", None}:
        if verbose:
            print(f"[plot_profile] save {out_png!r}")
        fig.savefig(out_png)


def cmd_plot(
    filename: str,
    kind: str,
    out_csv: Optional[str] = None,
    out_png: Optional[str] = None,
    title: Optional[str] = None,
    with_shape: bool = False,
    verbose: int = 0,
):
    """
    Plots a graph.

    :param filename: raw data to load
    :param kind: kind of plot to so, see below
    :param out_csv: output the data into that csv file
    :param out_png: output the graph in that file
    :param title: title (optional)
    :param with_shape: keep the shape to aggregate
    :param verbose: verbosity, if > 0, prints out the data in csv format

    Kinds of plots:

    * `'profile_op'`: draws the profiling per node type
    * `'profile_node'`: draws the profiling per node
    """
    if not os.path.exists(filename):
        raise FileNotFoundError(f"Unable to find {filename!r}.")
    allowed = {"profile_op", "profile_node"}
    if kind in allowed:
        plot_profile(filename, kind, out_csv, out_png, title=title, verbose=verbose)
    else:
        raise ValueError(f"Unexpected kind {kind:r}, it should be {allowed}.")



from typing import Any, Dict

import numpy as np

from onnx import NodeProto
from onnx.reference.op_run import OpRun
from .cpu.c_op_conv_ import ConvDouble, ConvFloat


class Conv(OpRun):
    def __init__(
        self, onnx_node: NodeProto, run_params: Dict[str, Any], schema: Any = None
    ):
        OpRun.__init__(self, onnx_node, run_params, schema)
        self.cache_: Dict[type, Any] = {}

    def _run(
        self,
        X,
        W,
        B=None,
        auto_pad=None,
        dilations=None,
        group=None,
        kernel_shape=None,
        pads=None,
        strides=None,
    ):
        if X.dtype not in self.cache_:
            if X.dtype == np.float32:
                rt = ConvFloat()
            elif X.dtype == np.float64:
                rt = ConvDouble()
            else:
                raise TypeError(
                    f"No C implementation C for operator 'Conv' and dtype={X.dtype}."
                )
            self.cache_[X.dtype] = rt

            rt.init(
                auto_pad,
                np.array(dilations or [], dtype=np.int64),
                group,
                np.array(kernel_shape or [], dtype=np.int64),
                np.array(pads or [], dtype=np.int64),
                np.array(strides or [], dtype=np.int64),
            )

        rt = self.cache_[X.dtype]

        if X is None:
            raise ValueError(f"X cannot be None for operator {type(self)}.")
        if min(X.shape) == 0:
            raise RuntimeError(
                f"Unable to run operator Conv on an empty matrix. X.shape={X.shape!r}."
            )
        if min(W.shape) == 0:
            raise RuntimeError(
                f"Unable to run operator Conv on an empty matrix. W.shape={W.shape!r}."
            )
        if B is not None and min(B.shape) == 0:
            raise RuntimeError(
                f"Unable to run operator Conv on an empty matrix. B.shape={B.shape!r}."
            )
        cv = rt.compute(X, W, B)
        return (cv,)



import os
import textwrap
from typing import List
from ... import _get_ort_ext_libs


def get_ort_ext_libs() -> List[str]:
    """
    Returns the list of libraries implementing new simple
    :epkg:`onnxruntime` kernels implemented for the
    :epkg:`CPUExecutionProvider`.
    """
    return _get_ort_ext_libs(os.path.dirname(__file__))


def documentation() -> List[str]:
    """
    Returns a list of rst string documenting every implemented kernels
    in this subfolder.
    """
    return list(
        map(
            textwrap.dedent,
            [
                """
    onnx_extented.ortops.option.cpu.TreeEnsembleClassifier
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    It does the sum of two tensors.

    **Provider**
    
    CPUExecutionProvider
    
    **Inputs**
    
    * X (T1): tensor of type T1

    **Outputs**

    * label (T3): labels of type T3
    * Y (T2): probabilities of type T2

    **Constraints**

    * T1: float, double
    * T2: float, double
    * T3: int64

    **Attributes**

    """,
                """
    onnx_extented.ortops.option.cpu.TreeEnsembleRegressor
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    It does the sum of two tensors.

    **Provider**
    
    CPUExecutionProvider
    
    **Inputs**
    
    * X (T1): tensor of type T1

    **Outputs**

    * Y (T2): prediction of type T2

    **Constraints**

    * T1: float, double
    * T2: float, double

    **Attributes**

    """,
            ],
        )
    )
